import math
from scipy.interpolate import interp1d
import numpy as np

from solcore.constants import kb, q, c, h
from solcore import si
from solcore.state import State

db_options = State()
db_options.db_mode = 'boltzmann'


def iv_detailed_balance(junction, options):
    """ Calculates the IV curve of a junction of kind "DB". This is a detailed balanced calculation of the IV curve of a PN junction characterized by a certain eqe, temperature and chemical potential (voltage). Normally, the eqe will be calculated based on a a given absorption edge and absorptance level resulting in a "top hat" type of eqe, but it can also be provided externally. By default, the solver uses the Boltzmann aproximation, although the full Planck equation might be used, also.

    :param junction: A Junction object of kind "DB"
    :param options: Other arguments for the calculation of the IV.
    :return:
    """
    T = options.T
    light = options.light_iv
    mode = options.db_mode
    Ta = options.T_ambient
    wl = options.wavelength

    try:
        Eg = junction.Eg
        n = junction.n

        if not hasattr(junction, 'eqe'):
            qe_detailed_balance(junction, wl)

        eqe = junction.eqe
        As = surface_integral(junction, wl)

        R_shunt = min(junction.R_shunt, 1e14) if hasattr(junction, 'R_shunt') else 1e14

    except AttributeError as err:
        raise AttributeError(
            'ERROR calculating the IV for the detailed balance Junction kind. Junction is missing one essential argument: {}'.format(
                err))

    j01 = 2 * q * n ** 2 * c
    # We loop over the (internal) voltages to get the (internal) current. The actual voltage applied is always smaller than Eg.
    junction.voltage = options.internal_voltages

    if mode == 'boltzmann':
        # If Boltzmann, the calculation can be vectorized as the voltages is outside the integral. No limit to V
        volt = junction.voltage

        def emis(v, T):
            boltz = j01 * (As / wl ** 4) * np.exp(- (h * c / wl) / (kb * T))
            junction.j01 = np.trapz(boltz, wl)
            out = np.trapz(boltz, wl) * np.exp((q * v) / (kb * T))
            return out

    else:
        # If the full Planck equation is used, we need to loop over the voltages. We limit them to the bandgap
        volt = np.where(junction.voltage < Eg - 0.001, junction.voltage, Eg - 0.001)

        def emis(v, T):
            out = []
            for vv in v:
                fermi = (As / wl ** 4) / (np.exp((h * c / wl - q * vv) / (kb * T)) - 1)
                out.append(np.trapz(fermi, wl))
            return j01*np.array(out)

    # Now we calculate the generation current, made of thermal generation and photogeneration
    jthermal = emis(np.array([0]), Ta)[0]
    if light:
        wl, ph = options.light_source.spectrum(x=wl, output_units='photon_flux_per_m')
        jsc = q * np.trapz(eqe(wl) * ph, wl) + jthermal
    else:
        jsc = jthermal

    junction.current = emis(volt, T) + volt / R_shunt - jsc

    junction.iv = interp1d(junction.voltage, junction.current, kind='linear', bounds_error=False, assume_sorted=True,
                           fill_value=(junction.current[0], junction.current[-1]))


def qe_detailed_balance(junction, wl):
    """ Calculates the EQE and the IQE in the detailed balanced limit... which is equal to the absorptance and the absorbed fraction of the light since, by definition of detailed balanced, the carrier collection is perfect.

    :param junction: Junction object of kind "DB"
    :param wl: wavelength in m
    :return: None
    """
    absorptance_detailed_balance(junction)

    # This is true just in the detailed balance limit
    junction.iqe = junction.absorptance

    try:
        z = np.linspace(0, junction.width, 1001)
        all_abs = junction.absorbed(z)
        abs_vs_wl = np.trapz(all_abs, z, axis=0)
        junction.eqe = interp1d(wl, abs_vs_wl, bounds_error=False, fill_value=(0, 0))

    except AttributeError:
        junction.eqe = junction.absorptance

    junction.qe = State({'WL': wl, 'IQE': junction.iqe(wl), 'EQE': junction.eqe(wl)})


def absorptance_detailed_balance(junction):
    """ Calculates the absorptance of the junction in the detailed balanced case. If it has not been calculated before, provided as input, it uses a "top hat" absorptance with the absorption edge at Eg and magnitude A.

    Note that while the junction.absorptance represents the potential fraction of light that can be absorbed at a given energy, the junction.absorbed represents the actual fraction of input light that is absorbed. For detailed balanced calculations, these two are equivalent to the IQE and EQE respectively because carrier collection is perfect, but that will not be the case, in general.

    :param junction: Junction object of kind "DB"
    """
    try:
        if not hasattr(junction, 'absorptance'):
            wlg = si(1240 / junction.Eg, 'nm')
            A = min(junction.A, 1)

            def absorptance(wl):
                return A * (wl < wlg)

            junction.absorptance = absorptance

    except AttributeError as err:
        raise AttributeError(
            'ERROR calculating absorptance for the detailed balance Junction kind.\nJunction is missing one essential argument: {}'.format(
                err))


def surface_integral(junction, wl):
    """ Calculates the surface integral of the absorptivity

    :param junction:
    :param options:
    :return:
    """
    Rn = np.minimum(0.999999, junction.reflected(wl))

    # Out of the normal incidence reflection we calculate an effective refractive index
    neff = (1 + np.sqrt(Rn)) / (1 - np.sqrt(Rn))

    # And the corresponding critical angle
    cos_tc = np.sqrt(1 - 1 / neff ** 2)
    steps = int(math.ceil(20 / (1 - min(cos_tc))))
    cos_t = np.linspace(0.001, 1, steps)

    # And now, we back calculate the reflection at any internal angle
    R = np.ones((len(wl), steps))

    for i in range(len(wl)):
        idx = np.where(cos_t > cos_tc[i])

        Rs = np.ones(steps)
        Rp = np.ones(steps)

        cos_i = np.sqrt(1 - neff[i] ** 2 * (1 - cos_t[idx] ** 2))

        Rs[idx] = ((cos_i - neff[i] * cos_t[idx]) / (cos_i + neff[i] * cos_t[idx])) ** 2
        Rp[idx] = ((cos_t[idx] - neff[i] * cos_i) / (cos_t[idx] + neff[i] * cos_i)) ** 2

        R[i, :] = (Rs + Rp) / 2

    # B = exp(-alpha*w) but it is not necessary to calculate alpha and w explicitly
    B = 1 - junction.absorptance(wl)
    A_back = 0.5 * (1 + R[:, -1] * B) * (1 - B)
    A_front = 0.5 * (1 - R[:, -1]) * (1 - B)

    for i in range(1, steps - 1):
        t = cos_t[i]

        # The integral of the absorpton/emission across the back surface
        A_back = A_back + (1 + R[:, i] * B ** (1 / t)) * (1 - B ** (1 / t)) * t

        # The integral of the absorpton/emission across the front surface
        A_front = A_front + (1 - R[:, i]) * (1 - B ** (1 / t)) * t

    A_back = A_back / (steps + 1)
    A_front = A_front / (steps + 1)

    return 2 * np.pi * (A_back + A_front)


if __name__ == '__main__':
    import numpy as np
    import matplotlib.pyplot as plt

    from solcore.state import State
    from solcore import material

    junction = State()
    options = State()

    wl = np.linspace(300, 900, 300) * 1e-9
    GaAs = material('GaAs')(T=300)

    n = GaAs.n(wl)
    cos_tc = np.sqrt(1 - 1 / n ** 2)
    Rn = ((n - 1) / (n + 1)) ** 2
    wlg = 800 * 1e-9
    A = 1

    V = 0
    T = 300

    options.wavelength = wl
    junction.reflected = interp1d(wl, Rn)
    junction.absorptance = lambda x: A * (x < wlg)

    #
    As = surface_integral(junction, options)

    jE = 2 * q * n ** 2 * c * (1 / wl ** 4) * As / (np.exp((h * c / wl - q * V) / kb / T) - 1)

    plt.plot(wl, jE)
    plt.show()


    #
    # all_cos_t = np.linspace(min(cos_tc), 1, 10)
    # Rs = np.zeros_like(n)
    # Rp = np.zeros_like(n)
    # Rnew = np.zeros_like(n)
    # for cos_t in all_cos_t:
    #     for i in range(len(wl)):
    #         if (cos_t > cos_tc[i]):
    #             cos_i = np.sqrt(1 - n[i] ** 2 * (1 - cos_t ** 2))
    #             Rs[i] = ((cos_i - n[i] * cos_t) / (cos_i + n[i] * cos_t)) ** 2
    #             Rp[i] = ((cos_t - n[i] * cos_i) / (cos_t + n[i] * cos_i)) ** 2
    #         else:
    #             Rs[i] = 1
    #             Rp[i] = 1
    #
    #         Rnew[i] = myR(wl[i], cos_t)
    #
    #     plt.plot(wl, (Rs + Rp) / 2, 'o')
    #     plt.plot(wl, Rnew, 'k')
    #
    # plt.show()



import os
from scipy.interpolate import interp1d
import numpy as np
from typing import Callable, Optional
from functools import wraps

from solcore.science_tracker import science_reference
from solcore import (
    spectral_conversion_nm_ev,
    spectral_conversion_nm_hz,
    eVnm,
    nmHz,
    nmJ,
)
from solcore.constants import q, h, c, kb

from solcore.light_source.spectral2 import (
    get_default_spectral2_object,
    calculate_spectrum_spectral2,
)
from solcore.light_source.smarts import (
    get_default_smarts_object,
    calculate_spectrum_smarts,
)


REGISTERED_CONVERTERS: dict = {}
""" Registered spectrum conversion functions."""


def reference_spectra():
    """ Function providing the standard reference spectra: AM0, AM1.5g and AM1.5d.

    :return: A 2D array with 4 columns representing the wavelength, AM0, AM1.5g and
    AM1.5d standard spectra.
    """

    science_reference(
        "Standard solar spectra",
        "ASTM G173-03(2012), Standard Tables for Reference Solar Spectral Irradiances: "
        "Direct Normal and Hemispherical on 37° Tilted Surface, ASTM International, "
        "West Conshohocken, PA, 2012, www.astm.org",
    )

    this_dir = os.path.split(__file__)[0]
    output = np.loadtxt(
        os.path.join(this_dir, "astmg173.csv"), dtype=float, delimiter=",", skiprows=2
    )

    return output


class LightSource:
    """ Common interface to access all types of light sources supported by Solcore.

    It includes standard solar spectra (AM0, AM1.5g, and AM1.5d), blackbody radiation,
    laser light or spectra created from atmospheric data using SPECTRAL2 or SMARTS.
    Additionally, it can also use experimentally measured spectra.
    """

    type_of_source = [
        "laser",
        "black body",
        "standard",
        "SMARTS",
        "SPECTRAL2",
        "custom",
    ]

    def __init__(self, source_type, x=None, output_units = "power_density_per_nm", concentration=1, **kwargs):
        """

        :param source_type:
        :param kwargs:
        """
        msg = f"Unknown source {source_type}. " \
              f"Valid options are: {self.type_of_source}"
        assert source_type in self.type_of_source, msg

        msg = f"Unknown output units {output_units}. " \
              f"Valid options are: {tuple(REGISTERED_CONVERTERS.keys())}"
        assert output_units in REGISTERED_CONVERTERS, msg

        self.source_type = source_type
        self.x = x
        self.x_internal = None
        self.power_density = 0

        self.options = {}
        self.output_units = output_units
        self.concentration = concentration
        self.options.update(kwargs)
        self._spectrum = None

        self._update_get_spectrum(self.output_units)
        self._update_spectrum_function()

        self.ready = False
        self.cache_spectrum = None

    def spectrum(self, x=None, output_units=None, concentration=None, **kwargs):
        """ Returns the spectrum of the light in the requested units. Internally,
        the spectrum is always managed in power density per nanometers, but the
        output can be provided in other compatible units, such as power density per
        Hertz or photon flux per eV.

        :param x: (Default=None) If "x" is provided, it must be an array with the
        spectral range in which to calculate the spectrum. Depending on the "units"
        defined when creating the light source, this array must be in nm, m, eV,
        J or hz.
        :param output_units: Units of the output spectrum
        :param concentration: Concentration of the light source
        :param kwargs: Options to update the light source. It can be "units",
        "concentration" or any of the options specific to the chosen type of light
        source.
        :return: Array with the spectrum in the requested units
        """

        if x is None:
            x = self.x if self.x is not None else self.x_internal

        output_units = output_units if output_units is not None else self.output_units
        con = concentration if concentration is not None else self.concentration

        self._update_get_spectrum(output_units)
        if kwargs:
            self._update(**kwargs)
            self._update_spectrum_function()
        return self.x, self._get_spectrum(self._spectrum, x) * con

    def _update(self, **kwargs):
        """ Updates the options of the light source with new values.

        It only updates existing options. No new options are added.

        :param kwargs: A dictionary with the options to update and their new values.
        :return: None
        """

        for opt in kwargs:
            if opt in self.options:
                self.options[opt] = kwargs[opt]

        self.ready = False
        self.cache_spectrum = None

    def _update_get_spectrum(self, output_units):
        """ Updates the function to get the spectrum, depending on the chosen output units.

        :return: None
        """
        msg = f"Valid units are: {list(REGISTERED_CONVERTERS.keys())}."
        assert output_units in REGISTERED_CONVERTERS, msg

        self._get_spectrum = REGISTERED_CONVERTERS[output_units]

    def _update_spectrum_function(self):
        """ Updates the spectrum function during the light source creation or just after updating one or more of the options. It also updates the "options" property with any default options available to the chosen light source, if any.

        :return: True
        """
        try:
            if self.source_type == "standard":
                self._spectrum = self._get_standard_spectrum(self.options)
            elif self.source_type == "laser":
                self._spectrum = self._get_laser_spectrum(self.options)
            elif self.source_type == "black body":
                self._spectrum = self._get_blackbody_spectrum(self.options)
            elif self.source_type == "SPECTRAL2":
                self._spectrum = self._get_spectral2_spectrum(self.options)
            elif self.source_type == "SMARTS":
                self._spectrum = self._get_smarts_spectrum(self.options)
            elif self.source_type == "custom":
                self._spectrum = self._get_custom_spectrum(self.options)
            else:
                raise ValueError(
                    "Unknown type of light source: {0}.\nValid light sources are: {1}".format(
                        self.source_type, self.type_of_source
                    )
                )

            self.ready = True

        except ValueError as err:
            print(err)

    def _get_standard_spectrum(self, options):
        """ Gets one of the reference standard spectra: AM0, AM1.5g or AM1.5d.

        :param options: A dictionary that contains the 'version' of the standard spectrum: 'AM0', 'AM1.5g' or 'AM1.5d'
        :return: A function that takes as input the wavelengths and return the standard spectrum at those wavelengths.
        """

        try:
            version = options["version"]

            spectra = reference_spectra()
            wl = spectra[:, 0]

            if version == "AM0":
                spectrum = spectra[:, 1]
            elif version == "AM1.5g":
                spectrum = spectra[:, 2]
            elif version == "AM1.5d":
                spectrum = spectra[:, 3]
            else:
                raise KeyError(
                    'ERROR when creating a standard light source. Input parameters must include "version" '
                    'which can be equal to "AM0", "AM1.5g" or "AM1.5d" only.'
                )

            self.x_internal = wl
            self.power_density = (
                np.trapz(y=spectrum, x=wl) * self.concentration
            )
            output = interp1d(
                x=wl, y=spectrum, bounds_error=False, fill_value=0, assume_sorted=True
            )
            return output

        except KeyError as err:
            print(err)

    def _get_laser_spectrum(self, options):
        """ Creates a gaussian light source with a given total power, linewidth and central wavelength. These three parameters must be provided in the "options" diccionary.

        :param options: A dictionary that must contain the 'power', the 'linewidth' and the 'center' of the laser emission.
        :return: A function that takes as input the wavelengths and return the laser spectrum at those wavelengths.
        """
        try:
            power = options["power"]
            sigma2 = options["linewidth"] ** 2
            center = options["center"]

            def output(x):
                out = (
                    power
                    / np.sqrt(2 * np.pi * sigma2)
                    * np.exp(-(x - center) ** 2 / 2 / sigma2)
                )
                return out

            self.x_internal = np.arange(
                center - 5 * options["linewidth"],
                center + 5 * options["linewidth"],
                options["linewidth"] / 20,
            )
            self.power_density = power * self.concentration
            return output

        except KeyError:
            print(
                'ERROR when creating a laser light source. Input parameters must include "power", "linewidth"'
                ' and "center".'
            )

    def _get_blackbody_spectrum(self, options):
        """ Gets the expontaneous emission in W/m2/sr/nm from a black body source chemical potential = 0

        :param options: A dictionary that must contain the temperature of the blackbody, in kelvin 'T' and the 'entendue' in sr. If not provided, the entendue will be taken as 1 sr. Possible values for entendue are:
            - 'Sun': The entendue will be taken as 6.8e-5 sr.
            - 'hemispheric': The entendue wil be taken as pi/2 sr.
            - A numeric value
        :return: A function that takes as input the wavelengths and return the black body spectrum at those wavelengths.
        """

        try:
            T = options["T"]
            if "entendue" in options:
                if options["entendue"] == "Sun":
                    entendue = 6.8e-5
                elif options["entendue"] == "hemispheric":
                    entendue = np.pi / 2
                else:
                    entendue = options["entendue"]
            else:
                entendue = 1
                options["entendue"] = 1

            def BB(x):
                x = x * 1e-9
                out = (
                    2
                    * entendue
                    * h
                    * c ** 2
                    / x ** 5
                    / (np.exp(h * c / (x * kb * T)) - 1)
                )
                return out * 1e-9

            wl_max = 2.897_772_9e6 / T
            self.x_internal = np.arange(0, wl_max * 10, wl_max / 100)
            sigma = 5.670_367e-8
            self.power_density = (
                sigma * T ** 4 * entendue / np.pi * self.concentration
            )

            return BB

        except KeyError:
            print(
                'ERROR when creating a blackbody light source. Input parameters must include "T" and, '
                'optionally, an "entendue", whose values can be "Sun" "hemispheric" or a number. '
                "Equal to 1 if omitted."
            )

    def _get_spectral2_spectrum(self, options):
        """ Get the solar spectrum calculated with the SPECTRAL2 calculator. The options dictionary is updated with the default options of all parameters if they are not provided.

        :param options: A dictionary that contain all the options for the calculator.
        :return: A function that takes as input the wavelengths and return the SPECTRAL2 calculated spectrum at those wavelengths.
        """
        default = get_default_spectral2_object()

        for opt in options:
            if opt in default:
                default[opt] = options[opt]

        options.update(default)

        wl, irradiance = calculate_spectrum_spectral2(options, power_density_in_nm=True)

        self.x_internal = wl
        self.power_density = (
            np.trapz(y=irradiance, x=wl) * self.concentration
        )
        output = interp1d(
            x=wl, y=irradiance, bounds_error=False, fill_value=0, assume_sorted=True
        )
        return output

    def _get_smarts_spectrum(self, options):
        """ Get the solar spectrum calculated with the SMARTS calculator. The options dictionary is updated with the default options of all parameters if they are not provided.

        :param options: A dictionary that contain all the options for the calculator.
        :return: A function that takes as input the wavelengths and return the SMARTS calculated spectrum at those wavelengths.
        """
        outputs = {
            "Extraterrestial": 2,
            "True direct": 2,
            "Experimental direct": 3,
            "Global horizontal": 4,
            "Global tilted": 5,
        }

        default = get_default_smarts_object()

        for opt in options:
            if opt in default:
                default[opt] = options[opt]

        options.update(default)

        if "output" not in options:
            options["output"] = "Global horizontal"

        try:
            output = outputs[options["output"]]

            out = calculate_spectrum_smarts(options)

            self.x_internal = out[0]
            self.power_density = (
                np.trapz(y=out[output], x=out[0]) * self.concentration
            )
            output = interp1d(
                x=out[0],
                y=out[output],
                bounds_error=False,
                fill_value=0,
                assume_sorted=True,
            )
            return output

        except KeyError:
            print(
                "ERROR: Output option no recognized. Avaliable options are: {}".format(
                    outputs
                )
            )
        except RuntimeError as err:
            print("ERROR in SMARTS: {}".format(err))

    def _get_custom_spectrum(self, options):
        """ Convert an custom spectrum into a solcore light source object.

        :param options: A dictionary that contains the following information:
            - 'x_data' and 'y_data' of the custom spectrum.
            - 'input_units', the units of the spectrum, such as 'photon_flux_per_nm' or 'power_density_per_eV'
        :return: A function that takes as input the wavelengths and return the custom spectrum at those wavelengths.
        """

        try:
            x_data = options["x_data"]
            y_data = options["y_data"]
            units = options["input_units"]

            # We check the units type by type.
            # Regardless of the input, we want power_density_per_nm
            if units == "power_density_per_nm":
                wl = x_data
                spectrum = y_data
            elif units == "photon_flux_per_nm":
                wl = x_data
                spectrum = y_data * (c * h * 1e9 / wl)
            elif units == "power_density_per_m":
                wl = x_data * 1e-9
                spectrum = y_data * 1e-9
            elif units == "photon_flux_per_m":
                wl = x_data * 1e-9
                spectrum = y_data * (c * h / wl)
            elif units == "power_density_per_eV":
                wl, spectrum = spectral_conversion_nm_ev(x_data, y_data)
            elif units == "photon_flux_per_eV":
                wl, spectrum = spectral_conversion_nm_ev(x_data, y_data)
                spectrum = spectrum * (c * h * 1e9 / wl)
            elif units == "power_density_per_J":
                wl, spectrum = spectral_conversion_nm_ev(x_data / q, y_data * q)
            elif units == "photon_flux_per_J":
                wl, spectrum = spectral_conversion_nm_ev(x_data / q, y_data * q)
                spectrum = spectrum * (c * h * 1e9 / wl)
            elif units == "power_density_per_hz":
                wl, spectrum = spectral_conversion_nm_hz(x_data, y_data)
            elif units == "photon_flux_per_hz":
                wl, spectrum = spectral_conversion_nm_hz(x_data, y_data)
                spectrum = spectrum * (h * x_data)
            else:
                raise ValueError(
                    "Unknown units: {0}.\nValid units are: {1}.".format(
                        units, list(REGISTERED_CONVERTERS.keys())
                    )
                )

            self.x_internal = wl
            self.power_density = (
                np.trapz(y=spectrum, x=wl) * self.concentration
            )
            output = interp1d(
                x=wl, y=spectrum, bounds_error=False, fill_value=0, assume_sorted=True
            )
            return output

        except KeyError as err:
            print(err)


def register_conversion_function(fun: Callable):
    """Registers a view method that will trigger an event. """

    @wraps(fun)
    def wrapper(*args, **kwargs):
        return fun(*args, **kwargs)

    REGISTERED_CONVERTERS[fun.__name__] = wrapper

    return wrapper


@register_conversion_function
def power_density_per_nm(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in power density per nanometer.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the wavelengths (in nm)
    :return: The spectrum in the chosen units.
    """
    return spectrum(x)


@register_conversion_function
def photon_flux_per_nm(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in photon flux per nanometer.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the wavelengths (in nm)
    :return: The spectrum in the chosen units.
    """
    return spectrum(x) / (c * h * 1e9 / x)


@register_conversion_function
def power_density_per_m(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in power density per meter.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the wavelengths (in m)
    :return: The spectrum in the chosen units.
    """
    return spectrum(x * 1e9) * 1e9


@register_conversion_function
def photon_flux_per_m(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in photon flux per meter.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the wavelengths (in m)
    :return: The spectrum in the chosen units.
    """
    return spectrum(x * 1e9) / (c * h / x) * 1e9


@register_conversion_function
def power_density_per_ev(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in power density per eV.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the energies (in eV)
    :return: The spectrum in the chosen units.
    """
    wavelength = eVnm(x)[::-1]
    output = spectrum(wavelength)
    _, output = spectral_conversion_nm_ev(wavelength, output)
    return output


@register_conversion_function
def photon_flux_per_ev(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in photon flux per eV.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the energies (in eV)
    :return: The spectrum in the chosen units.
    """
    wavelength = eVnm(x)[::-1]
    output = spectrum(wavelength)
    _, output = spectral_conversion_nm_ev(wavelength, output)
    return output / (q * x)


@register_conversion_function
def power_density_per_joule(
    spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray
):
    """ Function that returns the spectrum in power density per Joule.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the energies (in J)
    :return: The spectrum in the chosen units.
    """
    wavelength = nmJ(x)[::-1]
    output = spectrum(wavelength)
    _, output = spectral_conversion_nm_ev(wavelength, output)
    return output / q


@register_conversion_function
def photon_flux_per_joule(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in photon flux per Joule.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the energies (in J)
    :return: The spectrum in the chosen units.
    """
    wavelength = nmJ(x)[::-1]
    output = spectrum(wavelength)
    _, output = spectral_conversion_nm_ev(wavelength, output)
    return output / (q * x)


@register_conversion_function
def power_density_per_hz(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in power density per hertz.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the frequencies (in hz)
    :return: The spectrum in the chosen units.
    """
    wavelength = nmHz(x)[::-1]
    output = spectrum(wavelength)
    _, output = spectral_conversion_nm_hz(wavelength, output)
    return output


@register_conversion_function
def photon_flux_per_hz(spectrum: Callable[[np.ndarray], np.ndarray], x: np.ndarray):
    """ Function that returns the spectrum in photon flux per hertz.

    The input spectrum is assumed to be in power density per nanometer.

    :param spectrum: The spectrum to interpolate.
    :param x: Array with the frequencies (in hz)
    :return: The spectrum in the chosen units.
    """
    wavelength = nmHz(x)[::-1]
    output = spectrum(wavelength)
    frequency, output = spectral_conversion_nm_hz(wavelength, output)
    return output / (h * frequency)



""" Poisson_drift_diffusion related tests
"""
from pytest import approx


def test_light_iv(AlGaAs, light_source):
    import numpy as np
    from solcore.solar_cell import SolarCell, default_GaAs
    from solcore import material
    from solcore.solar_cell_solver import solar_cell_solver

    expected = np.array(
        [
            142.68025180227374,
            2.519346556870366,
            0.9169672186977382,
            329.61395441947565,
            2.347826086956522,
            140.3911287342211,
            0.3294918264376029,
        ]
    )

    T = AlGaAs.T
    Vin = np.linspace(-2, 2.61, 201)
    V = np.linspace(0, 2.6, 300)

    substrate = material("GaAs")(T=T)

    my_solar_cell = SolarCell(
        [AlGaAs, default_GaAs(T)], T=T, R_series=0, substrate=substrate
    )
    solar_cell_solver(
        my_solar_cell,
        "iv",
        user_options={
            "T_ambient": T,
            "db_mode": "boltzmann",
            "voltages": V,
            "light_iv": True,
            "wavelength": light_source.x,
            "optics_method": "BL",
            "mpp": True,
            "internal_voltages": Vin,
            "light_source": light_source,
        },
    )

    output = [
        my_solar_cell.iv.Isc,
        my_solar_cell.iv.Voc,
        my_solar_cell.iv.FF,
        my_solar_cell.iv.Pmpp,
        my_solar_cell.iv.Vmpp,
        my_solar_cell.iv.Impp,
        my_solar_cell.iv.Eta,
    ]

    assert np.array(output) == approx(expected, rel=1e-2)


def test_quantum_efficiency(AlGaAs, light_source):
    import numpy as np
    from solcore.solar_cell import SolarCell, default_GaAs
    from solcore import material
    from solcore.solar_cell_solver import solar_cell_solver

    expected = np.array(
        [
            0.9866334968497021,
            2.1512408472022467e-14,
            0.9779769012349702,
            0.03506561338387434,
        ]
    )
    T = AlGaAs.T
    Vin = np.linspace(-2, 2.61, 201)
    V = np.linspace(0, 2.6, 300)

    substrate = material("GaAs")(T=T)

    my_solar_cell = SolarCell(
        [AlGaAs, default_GaAs(T)], T=T, R_series=0, substrate=substrate
    )

    solar_cell_solver(
        my_solar_cell,
        "qe",
        user_options={
            "T_ambient": T,
            "db_mode": "boltzmann",
            "voltages": V,
            "light_iv": True,
            "wavelength": light_source.x,
            "optics_method": "BL",
            "mpp": True,
            "internal_voltages": Vin,
            "light_source": light_source,
        },
    )

    output = [
        my_solar_cell[0].eqe(500e-9),
        my_solar_cell[0].eqe(800e-9),
        my_solar_cell[1].eqe(700e-9),
        my_solar_cell[1].eqe(900e-9),
    ]
    assert np.array(output) == approx(expected, abs=1e-3)



from pytest import approx

import solcore


def test_constants_correctly_imported():
    q = 1.60217646e-19
    assert q == solcore.constants.q


def test_material_correctly_imported():
    GaAs = solcore.material("GaAs")(T=300)
    beta_Gamma_GaAs = 204
    assert GaAs.beta_Gamma == beta_Gamma_GaAs


def test_units_correctly_calculated():
    a_nm = 1239.8417166827828
    assert a_nm == approx(solcore.eVnm(1))



import numpy as np
from solcore import get_parameter
from solcore.constants import hbar, pi, c
from solcore.science_tracker import science_reference


def create_adachi_alpha(material, Esteps=(1.42, 6, 3000), T=300, wl=None):
    """ Calculates the n, k and absorption coefficient of a material using Adachi's formalism of critical points.

    :param material: A solcore material
    :param Esteps: (1.42, 6, 3000) A tuple with the start, end and step energies in which calculating the optical data
    :param T: (300) Temeprature in kelvin
    :param wl: (None) Optional array indicating the wavelengths in which calculating the data
    :return: A tuple containing 4 arrays: (Energy, n, k, alpha)
    """

    science_reference("Adachi optical dispersion relations",
                      'S.Adachi, “Optical dispersion relations for GaP, GaAs, GaSb, InP, InAs, InSb, AlxGa1−xAs, '
                      'and In1−xGaxAsyP1−y” J.Appl.Phys., vol.66, no.12, pp.6030–12, 1989.')

    if type(material) != str:
        material = material.plain_string()

    e0 = get_parameter(material, "E0", T=T) + 0j
    Delta0 = get_parameter(material, "E0plusD0", T=T) + 0j - e0
    e1 = get_parameter(material, "E1", T=T) + 0j
    Delta1 = get_parameter(material, "E1plusD1", T=T) + 0j - e1
    e2 = get_parameter(material, "E2", T=T) + 0j
    egid = get_parameter(material, "Eg1d", T=T) + 0j
    a = get_parameter(material, "A", T=T) + 0j
    b1 = get_parameter(material, "B1", T=T) + 0j
    b11 = get_parameter(material, "B11", T=T) + 0j
    Gamma = get_parameter(material, "Gamma", T=T) + 0j  # , "eV",T=T)
    cc = get_parameter(material, "C", T=T) + 0j
    gamma = get_parameter(material, "gamma", T=T) + 0j
    d = get_parameter(material, "D", T=T) + 0j
    omegaphonon = 0
    b2 = b1
    b21 = b11

    a0 = get_parameter(material, "lattice_constant", T=T)

    b2 = 44 * (e1 + 2 * Delta1 / 3.) / (a0 * (e1 + Delta1) ** 2)
    b1 = 44 * (e1 + Delta1 / 3) / (a0 * e1 ** 2)
    b2 = 0
    b1 = 7
    b11 = 2 * b1
    b21 = 2 * b2

    other_params = {
        "$e_0$": e0.real + 0.005,
        "$e_1$": e1.real,
        "$e_1+\Delta_1$": (e1 + Delta1).real,
        "$e_2$": e2.real,
        "$e_0+\Delta_0$": (e0 + Delta0).real,
        "$e_{g,1d}$": egid.real
    }
    # print b1, b2, b11,b21

    # print other_params
    if wl is not None:
        E = 1240 * 1e-9 / wl[::-1]
    else:
        E = np.linspace(*Esteps)  # electron Volts

    omega = E * 1.6e-19 / hbar

    re = lambda x: x.real
    H = lambda x: (x.real >= 0)
    i = 1j

    XiSO = E / (e0 + Delta0)
    Xi0 = E / e0

    f_Xi_0 = (Xi0) ** -2 * (2 - (1 + Xi0) ** 0.5 - (1 - Xi0) ** 0.5 * H(1 - Xi0))
    f_Xi_s0 = (XiSO) ** -2 * (2 - (1 + XiSO) ** 0.5 - (1 - XiSO) ** 0.5 * H(1 - XiSO))  #

    # equation 7
    epsilon_2_e0 = (a / E ** 2) * (
        (E - e0) ** 0.5 * H((Xi0) - 1) + 0.5 * (E - e0 - Delta0) ** 0.5 * H((E / (e0 + Delta0)) - 1))  #
    epsilon_1_e0 = a * e0 ** -1.5 * (f_Xi_0 + 0.5 * (e0 / (e0 + Delta0)) ** 1.5 * f_Xi_s0)

    Xi1 = E / e1  #
    Xi2 = E / e2  #
    Xi1s = E / (e1 + Delta1)  #

    # equation 12
    epsilon_2_e1 = pi * Xi1 ** -2 * (b1 - b11 * (e1 - E) ** 0.5 * H(1 - Xi1)) * H(
        re(b1 - b11 * (e1 - E) ** 0.5 * H(1 - Xi1)))  #
    epsilon_2_e1_d1 = pi * Xi1s ** -2 * (b2 - b21 * (e1 + Delta1 - E) ** 0.5 * H(
        1 - Xi1s))  # where is this from? #*H(b2-b21*(e1+Delta1-E)**0.5*H(1-Xi1s)) #
    # from numpy import sin
    #    epsilon_2_3dcp = pi*b1*Xi1**-2*H(Xi1-1) + pi*b2*Xi1s**-2*H(Xi1s-1) # +10*sin(30*E)#
    # epsilon_1_3dcp = -b1*(Xi1+i*Gamma/e1)**-2*log(1-(Xi1+i*Gamma/e1)**2) # Ned Version
    # epsilon_1_3dcp = -b1*Xi1**-2*log(1-Xi1**2)-b2*Xi1s**-2*log(1-Xi1s**2) #adachi version, no damping
    Gamma = 0.06
    E_damped = E + i * Gamma
    Xi1damped = E_damped / e1
    Xi1sdamped = E_damped / (e1 + Delta1)

    # equation 16
    epsilon_1_3dcp = -b1 * Xi1damped ** -2 * np.log(1 - Xi1damped ** 2) - b2 * Xi1sdamped ** -2 * np.log(
        1 - Xi1sdamped ** 2)  # adachi version
    epsilon_2_3dcp = pi * b1 * Xi1damped ** -2 * H(Xi1 - 1) + pi * b2 * Xi1sdamped ** -2 * H(Xi1s - 1)  # +10*sin(30*E)#

    # equation 16
    epsilon_2_e2 = cc * Xi2 * gamma / ((1 - Xi2 ** 2) ** 2 + (Xi2 * gamma) ** 2) * H(
        E - e1)  # added last term for compatibility with adachi's graph, dammit.
    # # print log(epsilon_2_e2)
    epsilon_1_e2 = cc * (1 - Xi2 ** 2) / ((1 - Xi2 ** 2) ** 2 + (Xi2 * gamma) ** 2)  #
    # print egid
    XiG_plus = (egid + hbar * omegaphonon) / E
    XiG_minus = (egid - hbar * omegaphonon) / E
    XiCh = E / e1

    epsilon_2_indirect = d / E ** 2 * (E - egid + hbar * omegaphonon) ** 2 * H(1 - XiG_plus) * H(1 - XiCh)  #
    epsilon_2_indirect_b = d / E ** 2 * (E - egid - hbar * omegaphonon) ** 2 * H(1 - XiG_minus) * H(1 - XiCh)  #
    # epsilon_2_indirect_minus = d/E**2 *(E-egid - hbar*omegaphonon)**2 * H(1-XiG_plus) * H(1-XiCh)# markus made this up a bit
    # epsilon_2_indirect_b_minus = d/E**2 *(E-egid + hbar*omegaphonon)**2 * H(1-XiG_minus) * H(1-XiCh)#markus made this up a bit

    epsilon_2_indirect_total = epsilon_2_indirect + epsilon_2_indirect_b

    eps1 = epsilon_1_e0 + epsilon_1_3dcp + epsilon_1_e2
    eps2 = epsilon_2_e0 + epsilon_2_e1 + epsilon_2_e1_d1 + epsilon_2_e2 + epsilon_2_indirect_total

    eps2 = epsilon_2_e0 + epsilon_2_e1 + epsilon_2_e2 + epsilon_2_indirect_total

    E = E * 1.6e-19
    k = np.sqrt((np.sqrt(eps1 ** 2 + eps2 ** 2) - eps1) / 2).real
    n = np.sqrt((np.sqrt(eps1 ** 2 + eps2 ** 2) + eps1) / 2).real
    alpha_data = 2 * omega / c * k

    # If the input was in wavelengths, we reverse the output arrays
    if wl is not None:
        alpha_data = alpha_data[::-1]
        n = n[::-1]
        k = k[::-1]

    return E.real, abs(n + 1e-10), abs(k + 1e-10), abs(alpha_data + 1e-10)


if __name__ == '__main__':
    from solcore import material
    from solcore.graphing import Graph, GraphData

    material_name = "InGaAs"
    solcore_material = material(material_name)(T=300, In=0.1)

    E, nn, kk, adachi_alpha_data = create_adachi_alpha(solcore_material, T=300)

    idx = np.argmin(abs(E-2))

    solcore_alpha_data = solcore_material.alphaE(E)

    E /= 1.6e-19

    idx = np.argmin(abs(E - 2))
    print(round(E[idx]*1000), round(nn[idx]*1000), round(kk[idx]*1000))

    out = [round(E[idx]*1000), round(nn[idx]*1000), round(kk[idx]*1000)]

    data = [2000.0, 3629.0, 308.0]

    print(data==out)

    curves = [
        GraphData(E, adachi_alpha_data / 100, label="adachi-Generated Data", linewidth=2, color="red"),
        GraphData(E, solcore_alpha_data / 100, label="Default Solcore Data", linewidth=2, color="grey"),
    ]

    g = Graph(curves, yscale="log", legend="best", xlim=(0, 6), ylim=(1000, 1e7)).draw()

    curves = [
        GraphData(E, nn, label="n", linewidth=2, color="red"),
        GraphData(E, kk, label="k", linewidth=2, color="grey"),
    ]

    g = Graph(curves, yscale="log", legend="best", xlim=(0, 6), ylim=(1e-3, 10)).draw()



# A number of custom functions/ classes than are designed to aid with plotting of graphs in Python.
# Good excuse to test out some OOP in python, T. Wilson 01-06-2016

# Custom Plot Colours :: This uses the colours and hex codes given online at;
# http://www.rapidtables.com/web/color/RGB_Color.htm
# to give access to a whole range of different colours.
# Will write as a class for practice.

import os
import re
from cycler import cycler, Cycler
from typing import List, Tuple, Dict, Union

# define the file path to the Colours.csv file, which sits within the Graphing Package
this_dir, this_fname = os.path.split(__file__)
file_path = os.path.join(this_dir, 'Colours.txt')

# Need to read the file in as text, declare
# opens .txt as file object
raw_txt: List[str] = []

with open(file_path) as file:

    raw_txt = file.read().splitlines()


def colours(colour: str, type: str = 'hex') -> Union[str, Tuple[float, ...]]:
    """colours(colour, type) :: Function identifies the name of the colour and returns the correct HEX or RGB key.
        Default return is in HEX.

    :param colour: string variable, name of colour as listed in the table of colours at;
                    http://www.rapidtables.com/web/color/RGB_Color.htm
    :param type: string variable, choices of "hex or "rgb". If colour is not found an error is returned.
    :return: HEX string or RGB tuple
    """

    # Initialise variables
    Cols: Dict[str, Union[str, Tuple[float, ...]]] = {}
    Cols.setdefault("Colour", colour)
    HEX2DEC: List[str] = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']

    for Lines in raw_txt:
        # Using regular expressions package, matching the material name with the correct section of the file.
        # If the name is found, ID is switched to True.
        split_line: List[str] = Lines.split(" = ")

        if re.fullmatch(colour.lower(), split_line[0]) != None:

            Cols["HEX"] = split_line[1]
            HEX: str = split_line[1]

            R: int = HEX2DEC.index(HEX[1]) * 16 ** 1 + HEX2DEC.index(HEX[2]) * 16 ** 0
            G: int = HEX2DEC.index(HEX[3]) * 16 ** 1 + HEX2DEC.index(HEX[4]) * 16 ** 0
            B: int = HEX2DEC.index(HEX[5]) * 16 ** 1 + HEX2DEC.index(HEX[6]) * 16 ** 0

            Cols["RGB"] = (R/255, G/255, B/255)

            break

        if "$END$" in Lines:
            raise ValueError("Colour name not found... Check http://www.rapidtables.com/web/color/RGB_Color.htm")

    if type.upper() == "HEX":
        return Cols["HEX"]

    elif type.upper() == "RGB":
        return Cols["RGB"]
    else:
        raise ValueError("Invalid 'type' selection... Choose either 'hex' or 'rgb'.")


def colour_cycle(name: str) -> Cycler:
    """colour_cycle(name, type) :: Function returns a cycler instance of a list of colours (in HEX) that can be used
        when plotting multi-plots.

    :param name: string variable matching the list of available colour cycles. Current list includes:
                    'rainbow', 'origin_basic', 'origin_system_colour_list'.
    :return: colour cycle instance
    """

    Arg_Out : List[Union[str, Tuple[float, ...]]] = []

    if name.lower() == 'rainbow': # 29 colours in shades of the rainbow...

        Arg_Out = [colours('Dark Red'), colours('Crimson'), colours('Salmon'),
                   colours('Orange Red'), colours('orange'), colours('Gold'),
                   colours('yellow'), colours('Lawn Green'), colours('Olive Drab'),
                   colours('Green', ), colours('Lime Green'), colours('Sea Green'),
                   colours('Teal'), colours('Cyan'), colours('Turquoise'),
                   colours('Steel Blue'), colours('Deep Sky Blue'), colours('Dodger Blue'),
                   colours('Blue'), colours('Navy'), colours('Indigo'),
                   colours('Slate Blue'), colours('Dark Magenta'), colours('Dark Orchid'),
                   colours('Purple'), colours('Magenta'), colours('Medium Violet Red'),
                   colours('Deep Pink'), colours('Black')]

    elif name.lower() == 'origin_basic': # 23 basic colours as defined in Origin...

        Arg_Out = ["#000000", "#FF0000", "#0000FF", "#FF00FF", "#008000", "#000080", "#8000FF", "#800080",
                   "#800000", "#808000", "#2B63A2", "#1E9696", "#9B641A", "#10C73E", "#B9247A", "#2DC5CC",
                   "#3F4198", "#93AC2B", "#808080", "#966464", "#649664", "#2BA3CA", "#326496"]

    elif name.lower() == "origin_system_color_list": # 13 colours defined by Origin's system colour list...

        Arg_Out = ["#F00820", "#FA3C3C", "#F08228", "#E6AF2D", "#E6DC32", "#A0E632", "#00DC00", "#00D28C",
                   "#00C8C8", "#00A0FF", "#1E3CFF", "#6E00DC", "#A000C8"]

    return cycler('color', Arg_Out)



# from scipy import interpolate as I
import numpy as np
from scipy.interpolate import make_interp_spline, splev
from typing import Union


class interp1d(object):
    """ Interpolate a 1D function.

    See Also
    --------
    splrep, splev - spline interpolation based on FITPACK
    UnivariateSpline - a more recent wrapper of the FITPACK routines
    """

    def __init__(self, x: np.ndarray, y: np.ndarray, kind: str = 'linear',
                 axis: int = -1, copy: bool = True, bounds_error: bool = False,
                 fill_value: float = np.nan):
        """ Initialize a 1D linear interpolation class.

        Description
        -----------
        x and y are arrays of values used to approximate some function f:
            y = f(x)
        This class returns a function whose call method uses linear
        interpolation to find the value of new points.

        Parameters
        ----------
        x : array
            A 1D array of monotonically increasing real values.  x cannot
            include duplicate values (otherwise f is overspecified)
        y : array
            An N-D array of real values.  y's length along the interpolation
            axis must be equal to the length of x.
        kind : str or int
            Specifies the kind of interpolation as a string ('linear',
            'nearest', 'zero', 'slinear', 'quadratic, 'cubic') or as an integer
            specifying the order of the spline interpolator to use.
        axis : int
            Specifies the axis of y along which to interpolate. Interpolation
            defaults to the last axis of y.
        copy : bool
            If True, the class makes internal copies of x and y.
            If False, references to x and y are used.
            The default is to copy.
        bounds_error : bool
            If True, an error is thrown any time interpolation is attempted on
            a value outside of the range of x (where extrapolation is
            necessary).
            If False, out of bounds values are assigned fill_value.
            By default, an error is raised.
        fill_value : float
            If provided, then this value will be used to fill in for requested
            points outside of the data range.
            If not provided, then the default is NaN.
        """

        self.copy = copy
        self.bounds_error = bounds_error
        self.fill_value = fill_value

        if kind in ['zero', 'slinear', 'quadratic', 'cubic']:
            order = {'nearest': 0, 'zero': 0, 'slinear': 1,
                     'quadratic': 2, 'cubic': 3}[kind]
            kind = 'spline'
        elif isinstance(kind, int):
            order = kind
            kind = 'spline'
        elif kind not in ('linear', 'nearest'):
            raise NotImplementedError("%s is unsupported: Use fitpack " \
                                      "routines for other types." % kind)
        x = np.array(x, copy=self.copy)
        y = np.array(y, copy=self.copy)

        if x.ndim != 1:
            raise ValueError("the x array must have exactly one dimension.")
        if y.ndim == 0:
            raise ValueError("the y array must have at least one dimension.")

        # Force-cast y to a floating-point type, if it's not yet one
        if not issubclass(y.dtype.type, np.inexact):
            y = y.astype(np.float_)

        # Normalize the axis to ensure that it is positive.
        self.axis = axis % len(y.shape)
        self._kind = kind

        if kind in ('linear', 'nearest'):
            # Make a "view" of the y array that is rotated to the interpolation
            # axis.
            axes = list(range(y.ndim))
            del axes[self.axis]
            axes.append(self.axis)
            oriented_y = y.transpose(axes)
            minval = 2
            len_y = oriented_y.shape[-1]
            if kind == 'linear':
                self._call = self._call_linear
            elif kind == 'nearest':
                self.x_bds = (x[1:] + x[:-1]) / 2.0
                self._call = self._call_nearest
        else:
            axes = list(range(y.ndim))
            del axes[self.axis]
            axes.insert(0, self.axis)
            oriented_y = y.transpose(axes)
            minval = order + 1
            len_y = oriented_y.shape[0]
            self._call = self._call_spline
            self._spline = make_interp_spline(x, oriented_y, k=order)

        len_x = len(x)
        if len_x != len_y:
            raise ValueError("x and y arrays must be equal in length along "
                             "interpolation axis.")
        if len_x < minval:
            raise ValueError("x and y arrays must have at "
                             "least %d entries" % minval)
        self.x = x
        self.y = oriented_y

    def _call_linear(self, x_new: np.ndarray) -> np.ndarray:

        # 2. Find where in the orignal data, the values to interpolate
        #    would be inserted.
        #    Note: If x_new[n] == x[m], then m is returned by searchsorted.
        x_new_indices = np.searchsorted(self.x, x_new)

        # 3. Clip x_new_indices so that they are within the range of
        #    self.x indices and at least 1.  Removes mis-interpolation
        #    of x_new[n] = x[0]
        x_new_indices = x_new_indices.clip(1, len(self.x) - 1).astype(int)

        # 4. Calculate the slope of regions that each x_new value falls in.
        lo = x_new_indices - 1
        hi = x_new_indices

        x_lo = self.x[lo]
        x_hi = self.x[hi]
        y_lo = self.y[..., lo]
        y_hi = self.y[..., hi]

        # Note that the following two expressions rely on the specifics of the
        # broadcasting semantics.
        slope = (y_hi - y_lo) / (x_hi - x_lo)

        # 5. Calculate the actual value for each entry in x_new.
        y_new = slope * (x_new - x_lo) + y_lo

        return y_new

    def _call_nearest(self, x_new: np.ndarray) -> np.ndarray:
        """ Find nearest neighbour interpolated y_new = f(x_new)."""

        # 2. Find where in the averaged data the values to interpolate
        #    would be inserted.
        #    Note: use side='left' (right) to searchsorted() to define the
        #    halfway point to be nearest to the left (right) neighbour
        x_new_indices = np.searchsorted(self.x_bds, x_new, side='left')

        # 3. Clip x_new_indices so that they are within the range of x indices.
        x_new_indices = x_new_indices.clip(0, len(self.x) - 1).astype(np.intp)

        # 4. Calculate the actual value for each entry in x_new.
        y_new = self.y[..., x_new_indices]

        return y_new

    def _call_spline(self, x_new: np.ndarray) -> np.ndarray:
        x_new = np.asarray(x_new)
        result = splev(x_new.ravel(), self._spline)
        return result.reshape(x_new.shape + result.shape[1:])

    def __call__(self, x_new: np.ndarray) -> np.ndarray:
        """Find interpolated y_new = f(x_new).

        Parameters
        ----------
        x_new : number or array
            New independent variable(s).

        Returns
        -------
        y_new : ndarray
            Interpolated value(s) corresponding to x_new.

        """

        # 1. Handle values in x_new that are outside of x.  Throw error,
        #    or return a list of mask array indicating the outofbounds values.
        #    The behavior is set by the bounds_error variable.
        x_new = np.asarray(x_new)
        out_of_bounds = self._check_bounds(x_new)

        y_new = self._call(x_new)

        # Rotate the values of y_new back so that they correspond to the
        # correct x_new values. For N-D x_new, take the last (for linear)
        # or first (for other splines) N axes
        # from y_new and insert them where self.axis was in the list of axes.
        nx = x_new.ndim
        ny = y_new.ndim

        # 6. Fill any values that were out of bounds with fill_value.
        # and
        # 7. Rotate the values back to their proper place.

        if nx == 0:
            # special case: x is a scalar
            if out_of_bounds:
                if ny == 0:
                    return np.asarray(self.fill_value)
                else:
                    y_new[...] = self.fill_value
            return np.asarray(y_new)
        elif self._kind in ('linear', 'nearest'):
            y_new[..., out_of_bounds] = self.fill_value
            axes = list(range(ny - nx))
            axes[self.axis:self.axis] = list(range(ny - nx, ny))
            return y_new.transpose(axes)
        else:
            y_new[out_of_bounds] = self.fill_value
            axes = list(range(nx, ny))
            axes[self.axis:self.axis] = list(range(nx))
            return y_new.transpose(axes)

    def _check_bounds(self, x_new: np.ndarray) -> np.ndarray:
        """Check the inputs for being in the bounds of the interpolated data.

        Parameters
        ----------
        x_new : array

        Returns
        -------
        out_of_bounds : bool array
            The mask on x_new of values that are out of the bounds.
        """

        # If self.bounds_error is True, we raise an error if any x_new values
        # fall outside the range of x.  Otherwise, we return an array indicating
        # which values are outside the boundary region.
        below_bounds = x_new < self.x[0]
        above_bounds = x_new > self.x[-1]

        # !! Could provide more information about which values are out of bounds
        if self.bounds_error and below_bounds.any():
            raise ValueError("A value in x_new is below the interpolation "
                             "range.")
        if self.bounds_error and above_bounds.any():
            raise ValueError("A value in x_new is above the interpolation "
                             "range.")

        # !! Should we emit a warning if some values are out of bounds?
        # !! matlab does not.
        out_of_bounds = np.logical_or(below_bounds, above_bounds)
        return out_of_bounds


class BilinearInterpolation(object):
    """docstring for BilinearInterpolation"""

    def __init__(self, x: Union[list, np.ndarray], y: Union[list, np.ndarray],
                 z: Union[list, np.ndarray], fill_value: float = 0.):
        super(BilinearInterpolation, self).__init__()
        self.x = np.array(x)
        self.y = np.array(y)
        self.z = np.array(z)

        # Check the shapes of x,y and z
        x_s = self.x.shape
        y_s = self.y.shape
        z_s = self.z.shape
        assert x_s[0] >= 3, "Need at least 3 points in x."
        assert y_s[0] >= 3, "Need at least 3 points in y."
        assert x_s[0] == z_s[
            0], "The number of rows in z (which is %s), must be the same as number of elements in x (which is %s)" % (
        x_s[0], z_s[0])
        assert y_s[0] == z_s[
            1], "The number of columns in z (which is %s), must be the same as number of elements in y (which is %s)" % (
        y_s[0], z_s[1])

    def __call__(self, xvalue: float, yvalue: float) -> np.ndarray:
        """The intepolated value of the surface."""
        try:
            x_i = 0
            found = False
            for i in list(range(0, len(self.x) - 1)):
                # print self.x[i], "<=", xvalue, "<", self.x[i+1]
                if self.x[i] <= xvalue < self.x[i + 1]:
                    # print "Found, returning index", i
                    x_i = i
                    found = True
                    break

            if not found:
                # print "Not found, returning 0."
                return 0.
            x_v = (self.x[x_i], self.x[x_i + 1])
            x_i = (x_i, x_i + 1)

            y_i = 0
            found = False
            for i in list(range(0, len(self.y) - 1)):
                # print self.y[i], "<=", yvalue, "<", self.y[i+1]
                if self.y[i] <= yvalue < self.y[i + 1]:
                    # print "Found, returning index", i
                    y_i = i
                    found = True
                    break

            if not found:
                # print "Not found, returning 0."
                return 0.
            y_v = (self.y[y_i], self.y[y_i + 1])
            y_i = (y_i, y_i + 1)

            D = (x_v[1] - x_v[0]) * (y_v[1] - y_v[0])
            ta = self.z[x_i[0], y_i[0]]
            tb = (x_v[1] - xvalue)
            tc = (y_v[1] - yvalue)
            t1 = ta * tb * tc / D
            # print t1, "=", ta, "*", tb, "*", tc, "/", D

            ta = self.z[x_i[1], y_i[0]]
            tb = (xvalue - x_v[0])
            tc = (y_v[1] - yvalue)
            t2 = ta * tb * tc / D
            # print t1, "=", ta, "*", tb, "*", tc, "/", D

            ta = self.z[x_i[0], y_i[1]]
            tb = (x_v[1] - xvalue)
            tc = (yvalue - y_v[0])
            t3 = ta * tb * tc / D
            # print t1, "=", ta, "*", tb, "*", tc, "/", D

            ta = self.z[x_i[1], y_i[1]]
            tb = (xvalue - x_v[0])
            tc = (yvalue - y_v[0])
            t4 = ta * tb * tc / D
            # print t1, "=", ta, "*", tb, "*", tc, "/", D

            # print "Therefore,"
            # print t1, "+", t2, "+", t3, "+", t4

            # print self.z
            return t1 + t2 + t3 + t4

        except ValueError:
            #            import pdb; pdb.set_trace()
            # Is one of the inputs an array?
            if type(xvalue) == list or type(xvalue) == tuple or type(xvalue) == type(np.array([])):
                constructed_list = []
                for sub_xvalue in xvalue:
                    constructed_list.append(self(sub_xvalue, yvalue))
                return np.array(constructed_list)


if __name__ == "__main__":
    # Some wavelength points
    x = np.array([400., 600., 601, 1000.])

    # Some angle values
    y = np.array([0., np.pi / 4, np.pi / 2])

    # Some reflectivity values
    #             rads @ 400nm,    rads @ 600nm,        rads @ 601m,             rads @ 1000nm
    z = np.array([[0., 0., 0.], [1e-9, 1e-9, 1e-9], [1 - 1e-9, 1 - 1e-9, 1 - 1e-9], [1., 1., 1.]])
    # print z.shape

    z_i = BilinearInterpolation(x, y, z)
    print(z_i(610, 0.1))



__author__ = 'diego'

""" Implementation of the mobility model by:

M. Sotoodeh, A. H. Khalid, and A. A. Rezazadeh,
“Empirical low-field mobility model for III–V compounds applicable in device simulation codes,”
J. Appl. Phys., vol. 87, no. 6, p. 2890, 2000.

"""

import json
import numpy as np
import os
from solcore.science_tracker import science_reference

# Constants
kb = 8.6173324e-5  # eV K-1
Log = lambda x: np.log10(x)

this_dir = os.path.split(__file__)[0]
parameters = os.path.join(this_dir, "mobility_parameters.json")
f = open(parameters, mode="r")
data = json.load(f)


def mobility_low_field(N, muMin, muMax, Nref, l, t1, t2, T=300):
    m = muMin + (muMax * (300 / T) ** t1 - muMin) / (1 + (N / (Nref * (300 / T) ** t2)) ** l)

    return m


def calculate_mobility(material, holes, N, x=0.0, y=0.0, T=300):
    """ Calculates the mobility using the model by Sotoodeh et al. If the material is not in the database, then the function returns the mobility for GaAs at that temperature, T, and impurity concentration, N.

    :param material: A string with the material name
    :param holes: If calculation should be done for electrons (holes=0) or holes (holes=1)
    :param N: Impurity concentration
    :param x: The fractional composition in the case of ternaries
    :param y: The other fractional composition in the case of quaternaries
    :param T: Temperature
    :return: The calculated mobility
    """
    science_reference('mobility calculator', 'M. Sotoodeh, A. H. Khalid, and A. A. Rezazadeh,'
                                             '“Empirical low-field mobility model for III–V compounds applicable in device simulation codes,"'
                                             'J. Appl. Phys., vol. 87, no. 6, p. 2890, 2000.')

    i = 1
    if holes: i = 2

    if material not in data.keys():
        print("Error: Material {0} not in the database for the mobility. Reverting to GaAs.".format(material))
        d = data['GaAs'][i]
    elif data[material][0] == 2:
        d = data[material][i]
    elif material == "InGaAs":
        d = calculate_InGaAs(x, i)
    elif material == "GaInP":
        d = calculate_InGaP(x, i, T)
    elif material == "AlGaAs":
        d = calculate_AlGaAs(x, i, T)
    elif material == "InAlAs":
        d = calculate_InAlAs(x, i, T)
    elif material == "InGaAsP":
        d = calculate_InGaAsP(x, y, i, T)
    else:
        d = calculate_General(material, x, i, T)

    muMin = d["muMin"]
    muMax = d["muMax"]
    Nref = d["Nref"]
    l = d["l"]
    t1 = d["t1"]
    t2 = d["t2"]

    m = mobility_low_field(N / 1e6, muMin, muMax, Nref, l, t1, t2, T) / 10000  # To convert it from cm2 to m2
    return m


def calculate_InGaAs(x, i):
    """ Calculates the parameters for an InGaAs alloy.

    :param x: Indium fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    p0 = data["InAs"][i]
    p1 = data["InGaAs"][i]
    p2 = data["GaAs"][i]

    xi = data["InGaAs"][3]["x"]

    newData = {}
    newData["muMin"] = interpolate_parameter_quad(x, p0["muMin"], p1["muMin"], p2["muMin"], xi)
    newData["muMax"] = interpolate_parameter_quad(x, p0["muMax"], p1["muMax"], p2["muMax"], xi)
    newData["Nref"] = 10 ** (interpolate_parameter_quad(x, Log(p0["Nref"]), Log(p1["Nref"]), Log(p2["Nref"]), xi))
    newData["l"] = interpolate_parameter_quad(x, p0["l"], p1["l"], p2["l"], xi)
    newData["t1"] = interpolate_parameter_quad(x, p0["t1"], p1["t1"], p2["t1"], xi)
    newData["t2"] = interpolate_parameter_quad(x, p0["t2"], p1["t2"], p2["t2"], xi)

    return newData


def calculate_InGaP(x, i, T):
    """ Calculates the parameters for an InGaP alloy.

    :param x: Indium fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    p0 = data["InP"][i]
    p1 = data["GaInP"][i]
    p2 = data["GaP"][i]

    xi = data["GaInP"][3]["x"]

    newData = {}
    newData["muMin"] = interpolate_parameter_quad(x, p0["muMin"], p1["muMin"], p2["muMin"], xi)
    newData["muMax"] = interpolate_parameter_quad(x, p0["muMax"], p1["muMax"], p2["muMax"], xi)
    newData["Nref"] = 10 ** (interpolate_parameter_quad(x, Log(p0["Nref"]), Log(p1["Nref"]), Log(p2["Nref"]), xi))
    newData["l"] = interpolate_parameter_quad(x, p0["l"], p1["l"], p2["l"], xi)
    newData["t1"] = interpolate_parameter_quad(x, p0["t1"], p1["t1"], p2["t1"], xi)
    newData["t2"] = interpolate_parameter_quad(x, p0["t2"], p1["t2"], p2["t2"], xi)

    if i == 1:
        # We have electrons, the above muMin and muMax need to be recalculated in an smarter way to take into
        # account the indirect bandgap above certain composition

        p0b = data["InP"][3]
        p1b = data["GaInP"][3]
        p2b = data["GaP"][3]

        # We calculate the band parameters for the alloy
        newBand = {}
        for key in p0b.keys():
            if key in ["es", "einf"]:
                newBand[key] = interpolate_epsilon(x, p0b[key], p2b[key])
            elif key in ["mnG", "mnX", "mnL"]:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key])
            else:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key], ABC=p1b["b{0}".format(key[2])])

        # Now we use these to calculate the direct and indirect mobilities (max and min)
        muDmax = (x - xi) / (1 - xi) * p0["muMax"] + (x - 1) / (xi - 1) * p1["muMax"]
        muDmin = (x - xi) / (1 - xi) * p0["muMin"] + (x - 1) / (xi - 1) * p1["muMin"]

        mind_alloy = mind(newBand["EgX"], newBand["EgL"], newBand["mnX"], newBand["mnL"], T)
        C = (p2b["mnX"] / mind_alloy) ** 1.5 * (p2b["einf"] * (-1) - p2b["es"] * (-1)) / (
        newBand["einf"] * (-1) - newBand["es"] * (-1))

        muImax = C * p2["muMax"]
        muImin = C * p2["muMin"]

        f = Rd(newBand["EgG"], newBand["EgX"], newBand["EgL"], newBand["mnG"], newBand["mnX"], newBand["mnL"], T)

        # Finally
        newData["muMin"] = f * muDmin + (1 - f) * muImin
        newData["muMax"] = f * muDmax + (1 - f) * muImax

    return newData


def calculate_AlGaAs(x, i, T):
    """ Calculates the parameters for an AlGaAs alloy.

    :param x: Al fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    p0 = data["AlAs"][i]
    p1 = data["AlGaAs"][i]
    p2 = data["GaAs"][i]

    xi = data["AlGaAs"][3]["x"]

    newData = {}
    if i == 2:
        # We have holes, which are easy: quadratic interpolation between AlAs, Al0.3GaAs and GaAs.
        newData["muMin"] = interpolate_parameter_quad(x, p0["muMin"], p1["muMin"], p2["muMin"], xi)
        newData["muMax"] = interpolate_parameter_quad(x, p0["muMax"], p1["muMax"], p2["muMax"], xi)
        newData["Nref"] = 10 ** (interpolate_parameter_quad(x, Log(p0["Nref"]), Log(p1["Nref"]), Log(p2["Nref"]), xi))
        newData["l"] = interpolate_parameter_quad(x, p0["l"], p1["l"], p2["l"], xi)
        newData["t1"] = interpolate_parameter_linear(x, p0["t1"], p2["t1"]) / (1 + x * (1 - x))
        newData["t2"] = p2["t2"]

    else:
        # We have electrons. Data is interpolated linearly between AlAs and GaAs
        newData["Nref"] = 10 ** (interpolate_parameter_linear(x, Log(p0["Nref"]), Log(p2["Nref"])))
        newData["l"] = interpolate_parameter_linear(x, p0["l"], p2["l"])
        newData["t1"] = interpolate_parameter_linear(x, p0["t1"], p2["t1"]) / (1 + x * (1 - x))
        newData["t2"] = interpolate_parameter_linear(x, p0["t2"], p2["t2"])

        # For muMin and muMax need to be recalculated in an smarter way to take into
        # account the indirect bandgap above certain composition

        p0b = data["AlAs"][3]
        p1b = data["AlGaAs"][3]
        p2b = data["GaAs"][3]

        # We calculate the band parameters for the alloy
        newBand = {}
        for key in p0b.keys():
            if key in ["es", "einf"]:
                newBand[key] = interpolate_epsilon(x, p0b[key], p2b[key])
            elif key in ["mnG", "mnX", "mnL"]:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key])
            else:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key], ABC=p1b["b{0}".format(key[2])])

        # Now we use these to calculate the direct and indirect mobilities (max and min)
        C = (p2b["mnG"] / newBand["mnG"]) ** 1.5 * (p2b["einf"] * (-1) - p2b["es"] * (-1)) / (
        newBand["einf"] * (-1) - newBand["es"] * (-1))

        muDmax = C * p2["muMax"]
        muDmin = C * p2["muMin"]

        muImax = p1["muMax"]
        muImin = p1["muMin"]

        f = Rd(newBand["EgG"], newBand["EgX"], newBand["EgL"], newBand["mnG"], newBand["mnX"], newBand["mnL"], T)

        # Finally
        newData["muMin"] = f * muDmin + (1 - f) * muImin
        newData["muMax"] = f * muDmax + (1 - f) * muImax

    return newData


def calculate_InAlAs(x, i, T):
    """ Calculates the parameters for an InAlAs alloy.

    :param x: Al fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    p0 = data["InAs"][i]
    p1 = data["InAlAs"][i]
    p2 = data["AlAs"][i]

    xi = data["InAlAs"][3]["x"]

    newData = {}
    if i == 2:
        # We have holes, which are easy
        newData["muMin"] = interpolate_parameter_linear(x, p0["muMin"], p2["muMin"]) / (1 + x * (1 - x))
        newData["muMax"] = interpolate_parameter_linear(x, p0["muMax"], p2["muMax"])
        newData["Nref"] = 10 ** (interpolate_parameter_linear(x, Log(p0["Nref"]), Log(p2["Nref"])))
        newData["l"] = interpolate_parameter_linear(x, p0["l"], p2["l"])
        newData["t1"] = interpolate_parameter_linear(x, p0["t1"], p2["t1"]) / (1 + x * (1 - x))
        newData["t2"] = interpolate_parameter_linear(x, p0["t2"], p2["t2"])

    else:
        # We have electrons
        newData["Nref"] = 10 ** (interpolate_parameter_quad(x, Log(p0["Nref"]), Log(p1["Nref"]), Log(p2["Nref"]), xi))
        newData["l"] = interpolate_parameter_quad(x, p0["l"], p1["l"], p2["l"], xi)
        newData["t1"] = interpolate_parameter_linear(x, p0["t1"], p2["t1"]) / (1 + x * (1 - x))
        newData["t2"] = interpolate_parameter_linear(x, p0["t2"], p2["t2"])

        # For muMin and muMax need to be recalculated in an smarter way to take into
        # account the indirect bandgap above certain composition

        p0b = data["InAs"][3]
        p1b = data["InAlAs"][3]
        p2b = data["AlAs"][3]

        # We calculate the band parameters for the alloy
        newBand = {}
        for key in p0b.keys():
            if key in ["es", "einf"]:
                newBand[key] = interpolate_epsilon(x, p0b[key], p2b[key])
            elif key in ["mnG", "mnX", "mnL"]:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key])
            else:
                newBand[key] = interpolate_parameter_linear(x, p0b[key], p2b[key], ABC=p1b["b{0}".format(key[2])])

        # Now we use these to calculate the direct and indirect mobilities (max and min)
        f1 = (x - xi) / (1 - xi)
        f2 = (x - 1) / (xi - 1)
        muDmax = p0["muMax"] ** f1 * p1["muMax"] ** f2
        muDmin = p0["muMin"] ** f1 * p1["muMin"] ** f2

        mind_alloy = mind(newBand["EgX"], newBand["EgL"], newBand["mnX"], newBand["mnL"], T)
        C = (p2b["mnX"] / mind_alloy) ** 1.5 * (p2b["einf"] * (-1) - p2b["es"] * (-1)) / (
        newBand["einf"] * (-1) - newBand["es"] * (-1))

        muImax = C * p2["muMax"]
        muImin = C * p2["muMin"]

        f = Rd(newBand["EgG"], newBand["EgX"], newBand["EgL"], newBand["mnG"], newBand["mnX"], newBand["mnL"], T)

        # Finally
        newData["muMin"] = f * muDmin + (1 - f) * muImin
        newData["muMax"] = f * muDmax + (1 - f) * muImax

    return newData


def calculate_InGaAsP(x, y, i, T):
    """ Calculates the parameters for an InGaAsP alloy. The calculation is based on a interpolation scheme between
    InGaP and InGaAs using data of compositions lattice matched to InP. Results for compositions away from this might
    not be very accurate.

    :param x: Indium fraction
    :param y: Phosphorus fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    p0 = calculate_InGaP(x, i, T)
    p1 = calculate_InGaAs(x, i)

    newData = {}
    newData["muMax"] = interpolate_parameter_linear(y, p0["muMax"], p1["muMax"]) / (1 + 6 * y * (1 - y))
    newData["Nref"] = 10 ** (interpolate_parameter_linear(y, Log(p0["Nref"]), Log(p1["Nref"])))
    newData["l"] = interpolate_parameter_linear(y, p0["l"], p1["l"])
    newData["t1"] = interpolate_parameter_linear(y, p0["t1"], p1["t1"]) / (1 + y * (1 - y))
    newData["t2"] = interpolate_parameter_linear(y, p0["t2"], p1["t2"])

    if i == 2:
        # We have holes
        newData["muMin"] = interpolate_parameter_linear(y, p0["muMin"], p1["muMin"])
    else:
        # We have electrons
        newData["muMin"] = interpolate_parameter_linear(y, p0["muMin"], p1["muMin"]) / (1 + 6 * y * (1 - y))

    return newData


def calculate_General(material, x, i, T):
    """ Calculates the parameters for a general alloy of the materials in the database assuming a simple linear
    interpolation. Only ternaries are supported this way.

    :param material: Material to calculate, which must be in the database
    :param x: Main fraction
    :param i: If the data for electrons (1) or holes (2) should be calculated
    :return:
    """

    parent1 = data[material][4]
    parent2 = data[material][5]

    p0 = data[parent1][i]
    p1 = data[parent2][i]

    newData = {}
    newData["muMin"] = interpolate_parameter_linear(x, p0["muMin"], p1["muMin"])
    newData["muMax"] = interpolate_parameter_linear(x, p0["muMax"], p1["muMax"])
    newData["Nref"] = 10 ** (interpolate_parameter_linear(x, Log(p0["Nref"]), Log(p1["Nref"])))
    newData["l"] = interpolate_parameter_linear(x, p0["l"], p1["l"])
    newData["t1"] = interpolate_parameter_linear(x, p0["t1"], p1["t1"])
    newData["t2"] = interpolate_parameter_linear(x, p0["t2"], p1["t2"])

    return newData


def Rd(EgG, EgX, EgL, mnG, mnX, mnL, T):
    """ Calculates the fraction of electrons in the direct valley.

    :param EgG: Gamma-valley bandgap
    :param EgX: X-valley bandgap
    :param EgL: L-valley bandgap
    :param mnG: Gamma-valley effective mass
    :param mnX: X-valley effective mass
    :param mnL: L-valley effective mass
    :param T: The temperature
    :return: The fraction.
    """

    RX = (mnX / mnG) ** 1.5 * np.exp((EgG - EgX) / (kb * T))
    RL = (mnL / mnG) ** 1.5 * np.exp((EgG - EgL) / (kb * T))

    fraction = 1. / (1 + RX + RL)
    return fraction


def mind(EgX, EgL, mnX, mnL, T):
    RX = np.exp((EgL - EgX) / (kb * T))

    fraction = 1. / (1 + RX)

    return mnL * fraction + mnX * (1 - fraction)


def interpolate_parameter_linear(x, AC, BC, ABC=0):
    return x * AC + (1 - x) * BC - ABC * x * (1 - x)


def interpolate_parameter_quad(x, y0, y1, y2, x1, x0=1, x2=0):
    dem0 = (x0 - x1) * (x0 - x2)
    dem1 = (x1 - x0) * (x1 - x2)
    dem2 = (x2 - x0) * (x2 - x1)

    nom0 = (x - x1) * (x - x2)
    nom1 = (x - x0) * (x - x2)
    nom2 = (x - x0) * (x - x1)

    p = y0 * nom0 / dem0 + y1 * nom1 / dem1 + y2 * nom2 / dem2

    return p


def interpolate_epsilon(x, AC, BC):
    ratioAC = (AC - 1) / (AC + 2)
    ratioBC = (BC - 1) / (BC + 2)

    k = interpolate_parameter_linear(x, ratioAC, ratioBC)

    return (2 * k + 1) / (1 - k)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    N = np.logspace(14, 21, 10)
    x = np.linspace(0, 1)
    # mo = calculate_mobility("InGaAsP", 0, N=1e17, x=x, y=0.5, T=300)
    mo = calculate_mobility("GaInP", 1, N=1e23, x=x, y=0.0, T=300)
    # mo = calculate_mobility("GaAs", 0, N=N, x=0, T=300)

    # plt.semilogx(N, mo)
    plt.plot(x, mo)
    plt.show()



"""
Calculates the k-points of the Brillouin zone in a given direction
"""
import numpy as np
from typing import List, Dict, Tuple

label_map: Dict[str, str] = {
    "Gamma": "$\Gamma$",
    "X": "X",
    "K": "K",
    "L": "L",
    "W": "W",
    "U": "U",
}


def brillouin_critical_points(a: float) -> Dict[str, np.ndarray]:
    return {
        "Gamma": np.array((0, 0, 0)),
        "X": np.pi / a * np.array((0, 2, 0)),
        "W": np.pi / a * np.array((1, 2, 0)),
        "L": np.pi / a * np.array((1, 1, 1)),
        "U": np.pi / a * np.array((1 / 2, 2, 1 / 2)),
        "K": np.pi / a * np.array((3 / 2, 3 / 2, 0))
    }


def traverse_brillouin(a: float,
                       traverse_order: tuple = ("L", "Gamma", "X",
                                                "W", "K", "Gamma"),
                       steps: int = 30) -> Tuple[np.ndarray, float,
                                                 List[Tuple[float, str]]]:

    critical_points = brillouin_critical_points(a)
    traverse_list = list(traverse_order)

    start_point = traverse_list.pop(0)

    k_coords = critical_points[start_point]

    coords = np.array([k_coords])

    xticks = [0]
    graph_cords = np.array([0])

    traverse_widths = []

    for i, target in enumerate(traverse_list):
        l_vec = (critical_points[target] - k_coords) * a
        traverse_widths.append(np.sqrt(sum(l_vec * l_vec)))
        k_coords = critical_points[target]

    k_coords = critical_points[start_point]

    for i, target in enumerate(traverse_list):
        coords = np.vstack(
            (coords, k_coords + np.matrix(np.linspace(0, 1, steps)[1:]).transpose() * (critical_points[target] - k_coords)))
        graph_cords = np.hstack((graph_cords, np.linspace(graph_cords[-1], graph_cords[-1] + traverse_widths[i], steps)[1:]))
        xticks.append(graph_cords[-1])
        k_coords = critical_points[target]

    scale = max(graph_cords)
    return np.array(coords), graph_cords / scale, list(zip(xticks / scale, [label_map[s] for s in traverse_order]))


def kvector(a: float, t: float = 0, p: float = np.pi,
            fraction: float = 0.2, points: int = 50,
            vin: np.ndarray = None) -> np.ndarray:
    """ Calculates the k points in a direction given by the spheric angles theta (t) and phi (p).

    The fraction of the Brilluin zone calculated is given by "fraction" and the number of points by "points".
    If "vin" is given, the direction of interest is taken from this vector."""

    s3 = np.sqrt(3)

    # Maximum k in the high symmetry directions
    Xkmax = 2 * np.pi / a
    Lkmax = s3 * np.pi / a

    # Unity vectors of the high symmetry directions (normalised)
    X = np.array([[1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, -1, 0], [0, 0, 1], [0, 0, -1]])
    L = 1 / s3 * np.array(
        [[1, 1, 1], [1, 1, -1], [1, -1, 1], [1, -1, -1], [-1, 1, 1], [-1, 1, -1], [-1, -1, -1], [-1, -1, 1]])

    # Unity vector in the direction of interest
    if vin is not None:
        u = vin / np.linalg.norm(vin)
    else:
        u = np.array([np.cos(t) * np.sin(p), np.sin(t) * np.sin(p), np.cos(p)])

    # We calculate the angle between u and the X and L vectors. The one producing the smallest norm will give the bounding plane.
    Xcos = Xkmax / max(abs(np.dot(X, u)))
    Lcos = Lkmax / max(abs(np.dot(L, u)))

    kmax = min(Xcos, Lcos)

    # Finally, we create the kx, ky and kz values in the u direction with the maximum norm kmax.
    magnitudes = kmax * np.linspace(0, fraction, points)
    output = []

    for i in range(len(magnitudes)):
        output.append(magnitudes[i] * u)

    return np.array(output)



from solcore.constants import *
from numpy import array
import numpy as np
from solcore.science_tracker import science_reference

H = lambda x: 0 if x < 0 else 1  ### blindly changed this !!
D = lambda x, width: 1 if (x >= -width) and (x <= width) else 0  # Delta function
L = lambda x, centre, hwhm: 1 / pi * (0.5 * hwhm) / (
    (x - centre) ** 2 + (0.5 * hwhm) ** 2)  # Lorenzian (area normalised to 1)
Gauss = lambda x, centre, hwhm: 1 / np.sqrt(2 * pi) / (0.5 * hwhm) * np.exp(
    -0.5 * (x - centre) ** 2 / (0.5 * hwhm) ** 2)  # Gaussian (area normalised to 1)


def exciton_rydberg_energy_2d(me, mh, eps_r):
    """
    :param me: electron effective mass (units: kg)
    :param mh: hole effective mass (units: kg)
    :param eps_r: dielectic constant (units: SI)
    :return: The exciton Rydberg energy
    """

    science_reference("Definition of the exciton Rydberg Energy for a quantum well.",
                      "S. L. Chuang, Physics of Optoelectonic Devices, Second Edition, p.554, Table 13.1")

    mr = me * mh / (me + mh)  # reduced effective mass
    if True:
        return mr * q ** 4 / (2 * hbar ** 2 * (4 * np.pi * eps_r * vacuum_permittivity) ** 2)
    Ry = 13.6 * 1.6e-19
    m0 = electron_mass
    Ry_eff = (mr / m0) * Ry / eps_r ** 2  # Effective Rydberg
    return Ry_eff


def exciton_bohr_radius(me, mh, eps):
    """
    :param me: electron effective mass (units: kg)
    :param mh: hole effective mass (units: kg)
    :param eps: dielectic constant (units: SI)
    :return: Exciton Borh radius
    """

    science_reference("Definition of the exciton bohr radius for a quantum well.",
                      "S. L. Chuang, Physics of Optoelectonic Devices, Second Edition, p.554, Table 13.1")
    mr = me * mh / (me + mh)  # reduced effective mass
    return hbar ** 2 / mr * (4 * np.pi * eps / q ** 2)


def alpha_c_hh_TE(E, z, E_e, E_hh, psi_e, psi_hh, well_width, me, mhh, Ep, nr):
    """     Absortion coefficient for incident light forming a transision between hh and c band of a quantum well

    NB. Assumes that valence band is a zero energy, might need to manually apply an offset.

    :param E: photon energy (units: J)
    :param z: mesh points along growth direction (units: m)
    :param E_e: Electron state energy (units: J)
    :param E_hh: Heavy hole state energy (units: J)
    :param psi_e: Electron envelope function (psi_e^2 must be normalised)
    :param psi_hh: Heavy hole envelope function (psi_hh^2 must be normalised)
    :param well_width: (units: m)
    :param me: electron effective mass (units: kg)
    :param mhh: heavy hole effective mass (units: kg)
    :param Ep: the Kane parameter "Optical dipole matrix elemet", sometimes "Momentum matrix element", e.g. Ep for GaAs ~ 28eV
    :param nr: refractive index
    :return:
    """

    m0 = electron_mass
    C0 = np.pi * q ** 2 / (nr * c * vacuum_permittivity * m0 ** 2 * E / hbar)
    mr = me * mhh / (me + mhh)  # reduced effective mass
    DOS_2D = mr / (np.pi * hbar ** 2 * well_width)
    Mbsq_3D = m0 / 6 * Ep
    Mbsq_2D = 3 / 2 * Mbsq_3D
    Ieh = np.trapz(psi_e * psi_hh, z) ** 2
    return C0 * Ieh * Mbsq_2D * DOS_2D * H(E - (E_e - E_hh))


def alpha_c_lh_TE(E, z, E_e, E_lh, psi_e, psi_lh, well_width, me, mlh, Ep, nr):
    """     Absortion coefficient for incident light forming a transision between hh and c band of a quantum well
    
    NB. Assumes that valence band is a zero energy, might need to manually apply an offset.
    
    :param E: photon energy (units: J)
    :param z: mesh points along growth direction (units: m)
    :param E_e: Electron state energy (units: J)
    :param E_hh: Heavy hole state energy (units: J)
    :param psi_e: Electron envelope function (psi_e^2 must be normalised)
    :param psi_hh: Heavy hole envelope function (psi_hh^2 must be normalised)
    :param well_width: (units: m)
    :param me: electron effective mass (units: kg)
    :param mhh: heavy hole effective mass (units: kg)
    :param Ep: the Kane parameter "Optical dipole matrix elemet", sometimes "Momentum matrix element", e.g. Ep for GaAs ~ 28eV
    :param nr: refractive index
    :return:
    """
    m0 = electron_mass
    C0 = np.pi * q ** 2 / (nr * c * vacuum_permittivity * m0 ** 2 * E / hbar)
    mr = me * mlh / (me + mlh)  # reduced effective mass
    DOS_2D = mr / (np.pi * hbar ** 2 * well_width)
    Mbsq_3D = m0 / 6 * Ep
    Mbsq_2D = 1 / 2 * Mbsq_3D
    Ieh = np.trapz(psi_e * psi_lh, z) ** 2
    return C0 * Ieh * Mbsq_2D * DOS_2D * H(E - (E_e - E_lh))


def alpha_exciton_ehh_TE(exciton_index, E, z, E_e, E_hh, psi_e, psi_hh, well_width, me, mhh, Ep, nr, eps,
                         hwhm=6e-3 * 1.6e-19, dimensionality=0.15, line_shape="Lorenzian"):
    """

    :param exciton_index:
    :param E:
    :param z:
    :param E_e:
    :param E_hh:
    :param psi_e:
    :param psi_hh:
    :param well_width:
    :param me:
    :param mhh:
    :param Ep:
    :param nr:
    :param eps:
    :param hwhm:
    :param dimensionality:
    :param line_shape:
    :return:
    """
    if exciton_index < 1:
        raise ValueError("The excition index must start on 1.")

    if dimensionality < 0 or dimensionality > 0.5:
        raise ValueError("The dimensionality parameter of the exciton must be a number between 0 and 0.5 (inclusive).")

    # Borrow constants from the bound-to-bound calculation
    m0 = electron_mass
    C0 = np.pi * q ** 2 / (nr * c * vacuum_permittivity * m0 ** 2 * E / hbar)
    mr = me * mhh / (me + mhh)  # reduced effective mass
    DOS_2D = mr / (np.pi * hbar ** 2 * well_width)
    Mbsq_3D = m0 / 6 * Ep
    Mbsq_2D = 3 / 2 * Mbsq_3D
    Ieh = np.trapz(psi_e * psi_hh, z) ** 2

    # Excitions are considered by modiftying the bulk absorption coefficient using an oscillator strength
    Ry_eff = exciton_rydberg_energy_2d(me, mhh, eps / vacuum_permittivity)
    En = -Ry_eff / ((exciton_index - dimensionality) ** 2)  # Exciton binding energy
    Et = (E_e - E_hh)
    # import pdb; pdb.set_trace()
    # NOTE TO MARKUS: Added the possiblity of using a Gauss lineshape
    if line_shape is "Gauss":
        shape = Gauss(E, Et + En, hwhm)
    else:
        shape = L(E, Et + En, hwhm)
    # NOTE TO MARKUS: It seems that there was a factor 2 missing in the oscilator strength as well as the exciton index. 
    # See: P C Klipstein and N Apsley 1986 J. Phys. C: Solid State Phys. 19 6461 doi:10.1088/0022-3719/19/32/020 
    oscillator_strength = (2 * Ry_eff / (exciton_index - dimensionality) ** 3)
    return C0 * Ieh * Mbsq_2D * DOS_2D * oscillator_strength * shape


def alpha_exciton_elh_TE(exciton_index, E, z, E_e, E_lh, psi_e, psi_lh, well_width, me, mlh, Ep, nr, eps,
                         hwhm=6e-3 * 1.6e-19, dimensionality=0.15, line_shape="Lorenzian"):
    """

    :param exciton_index:
    :param E:
    :param z:
    :param E_e:
    :param E_lh:
    :param psi_e:
    :param psi_lh:
    :param well_width:
    :param me:
    :param mlh:
    :param Ep:
    :param nr:
    :param eps:
    :param hwhm:
    :param dimensionality:
    :param line_shape:
    :return:
    """
    if exciton_index < 1:
        raise ValueError("The excition index must start on 1.")

    if dimensionality < 0 or dimensionality > 0.5:
        raise ValueError("The dimensionality parameter of the exciton must be a number between 0 and 0.5 (inclusive).")

    # Borrow constants from the bound-to-bound calculation
    m0 = electron_mass
    C0 = np.pi * q ** 2 / (nr * c * vacuum_permittivity * m0 ** 2 * E / hbar)
    mr = me * mlh / (me + mlh)  # reduced effective mass
    DOS_2D = mr / (np.pi * hbar ** 2 * well_width)
    Mbsq_3D = m0 / 6 * Ep
    Mbsq_2D = 1 / 2 * Mbsq_3D
    Ieh = np.trapz(psi_e * psi_lh, z) ** 2

    # Excitions are considered by modiftying the bulk absorption coefficient using an oscillator strength
    Ry_eff = exciton_rydberg_energy_2d(me, mlh, eps / vacuum_permittivity)
    En = -Ry_eff / ((exciton_index - dimensionality) ** 2)  # Exciton binding energy
    Et = (E_e - E_lh)
    # NOTE TO MARKUS: Added the possiblity of using a Gauss lineshape
    if line_shape is "Gauss":
        shape = Gauss(E, Et + En, hwhm)
    else:
        shape = L(E, Et + En, hwhm)
    # NOTE TO MARKUS: It seems that there was a factor 2 missing in the oscilator strength as well as the exciton index.
    # See: P C Klipstein and N Apsley 1986 J. Phys. C: Solid State Phys. 19 6461 doi:10.1088/0022-3719/19/32/020
    oscillator_strength = (2 * Ry_eff / (exciton_index - dimensionality) ** 3)
    return C0 * Ieh * Mbsq_2D * DOS_2D * oscillator_strength * shape


def sum_alpha_c_hh_TE(E, z, E_e, E_hh, psi_e, psi_hh, well_width, me, mh, Ep, nr):
    """

    :param E:
    :param z:
    :param E_e:
    :param E_hh:
    :param psi_e:
    :param psi_hh:
    :param well_width:
    :param me:
    :param mh:
    :param Ep:
    :param nr:
    :return:
    """
    alpha = 0
    for ee, pe in zip(E_e, psi_e):
        for eh, ph in zip(E_hh, psi_hh):
            alpha += alpha_c_hh_TE(E, z, ee, eh, pe, ph, well_width, me, mh, Ep, nr)
    return alpha


def sum_alpha_c_lh_TE(E, z, E_e, E_lh, psi_e, psi_lh, well_width, me, mh, Ep, nr):
    """

    :param E:
    :param z:
    :param E_e:
    :param E_lh:
    :param psi_e:
    :param psi_lh:
    :param well_width:
    :param me:
    :param mh:
    :param Ep:
    :param nr:
    :return:
    """
    alpha = 0
    for ee, pe in zip(E_e, psi_e):
        for eh, ph in zip(E_lh, psi_lh):
            alpha += alpha_c_lh_TE(E, z, ee, eh, pe, ph, well_width, me, mh, Ep, nr)
    return alpha


def sum_alpha_exciton_c_hh_TE(E, z, E_e, E_hh, psi_e, psi_hh, well_width, me, mh, Ep, nr, eps, hwhm=6e-3 * 1.6e-19,
                              dimensionality=0.5, line_shape="Lorenzian"):
    """

    :param E:
    :param z:
    :param E_e:
    :param E_hh:
    :param psi_e:
    :param psi_hh:
    :param well_width:
    :param me:
    :param mh:
    :param Ep:
    :param nr:
    :param eps:
    :param hwhm:
    :param dimensionality:
    :param line_shape:
    :return:
    """
    alpha = 0
    eh_pairs_with_zero_angular_momentum = zip(range(len(E_e)),
                                              range(len(E_hh)))  # l=0 for optically allowed transitions
    for exciton_index, eh_quantum_numbers in enumerate(eh_pairs_with_zero_angular_momentum):
        e_index = eh_quantum_numbers[0]
        hh_index = eh_quantum_numbers[1]
        ee = E_e[e_index]
        ehh = E_hh[hh_index]
        pe = psi_e[e_index]
        phh = psi_hh[e_index]
        alpha += alpha_exciton_ehh_TE(exciton_index + 1, E, z, ee, ehh, pe, phh, well_width, me, mh, Ep, nr, eps,
                                      hwhm=hwhm, dimensionality=dimensionality, line_shape=line_shape)
    return alpha


def sum_alpha_exciton_c_lh_TE(E, z, E_e, E_lh, psi_e, psi_lh, well_width, me, mlh, Ep, nr, eps, hwhm=6e-3 * 1.6e-19,
                              dimensionality=0.5, line_shape="Lorenzian"):
    """

    :param E:
    :param z:
    :param E_e:
    :param E_lh:
    :param psi_e:
    :param psi_lh:
    :param well_width:
    :param me:
    :param mlh:
    :param Ep:
    :param nr:
    :param eps:
    :param hwhm:
    :param dimensionality:
    :param line_shape:
    :return:
    """
    alpha = 0
    eh_pairs_with_zero_angular_momentum = zip(range(len(E_e)),
                                              range(len(E_lh)))  # l=0 for optically allowed transitions
    for exciton_index, eh_quantum_numbers in enumerate(eh_pairs_with_zero_angular_momentum):
        e_index = eh_quantum_numbers[0]
        lh_index = eh_quantum_numbers[1]
        ee = E_e[e_index]
        elh = E_lh[lh_index]
        pe = psi_e[e_index]
        plh = psi_lh[e_index]
        alpha += alpha_exciton_elh_TE(exciton_index + 1, E, z, ee, elh, pe, plh, well_width, me, mlh, Ep, nr, eps,
                                      hwhm=hwhm, dimensionality=dimensionality, line_shape=line_shape)
    return alpha


def calc_alpha(QM_result, well_width, kane_parameter=28 * 1.6e-19, refractive_index=3.5, hwhm=6e-3 * 1.6e-19,
               dimensionality=0.5, theta=0, eps=12.9 * vacuum_permittivity, espace=None, line_shape="Lorenzian"):
    """ Calculates the absorption coeficient of a quantum well structure assuming the parabolic approximation for the
    effective masses.

    :param QM_result: The output of the Schrodinger solver, incldued in the 'quantum_mechanics' package
    :param well_width: The well width
    :param kane_parameter: The Kane parameter
    :param refractive_index: Refractive (effective) index of the QW
    :param hwhm: Full width at half maximum of the excitonic lineshape
    :param dimensionality:
    :param theta:
    :param eps:
    :param espace:
    :param line_shape:
    :return:
    """
    results = QM_result

    # We pick the effective mass at the center of the structure as THE efective mass of the QW
    points = len(results["effective_masses"]["me"])
    me = results["effective_masses"]["me"][round(points / 2.0)]
    try:
        mhh = results["effective_masses"]["mhh"][round(points / 2.0)]
        mlh = results["effective_masses"]["mlh"][round(points / 2.0)]
    except KeyError:
        # If the calculation mode for the bands is kp4x4 or kp6x6, we have to use the effective mass transverse to the
        # growth direction (in the QW plane).
        mhh = results["effective_masses"]["mhh_t"][round(points / 2.0)]
        mlh = results["effective_masses"]["mlh_t"][round(points / 2.0)]
    #    print (results["effective_masses"]["me"][round(points/2.0)])
    #    print (results["effective_masses"]["mhh"][round(points/2.0)])
    #    print (results["effective_masses"]["mlh"][round(points/2.0)])

    result_c_hh_TE = []
    result_c_lh_TE = []
    result_ex_c_hh_TE = []
    result_ex_c_lh_TE = []
    for e in espace:
        c_hh = sum_alpha_c_hh_TE(e, results['x'], results['E']['Ee'], results['E']['Ehh'],
                                 results["wavefunctions"]['psi_e'], results["wavefunctions"]['psi_hh'], well_width, me,
                                 mhh, kane_parameter, refractive_index)
        c_lh = sum_alpha_c_lh_TE(e, results['x'], results['E']['Ee'], results['E']['Elh'],
                                 results["wavefunctions"]['psi_e'], results["wavefunctions"]['psi_lh'], well_width, me,
                                 mlh, kane_parameter, refractive_index)
        ex_c_hh = sum_alpha_exciton_c_hh_TE(e, results['x'], results['E']['Ee'], results['E']['Ehh'],
                                            results["wavefunctions"]['psi_e'], results["wavefunctions"]['psi_hh'],
                                            well_width, me, mhh, kane_parameter, refractive_index, eps, hwhm=hwhm,
                                            dimensionality=dimensionality, line_shape=line_shape)
        ex_c_lh = sum_alpha_exciton_c_lh_TE(e, results['x'], results['E']['Ee'], results['E']['Elh'],
                                            results["wavefunctions"]['psi_e'], results["wavefunctions"]['psi_lh'],
                                            well_width, me, mlh, kane_parameter, refractive_index, eps, hwhm=hwhm,
                                            dimensionality=dimensionality, line_shape=line_shape)
        result_c_hh_TE.append(c_hh)
        result_c_lh_TE.append(c_lh)
        result_ex_c_hh_TE.append(ex_c_hh)
        result_ex_c_lh_TE.append(ex_c_lh)

    result_c_hh_TE = array(result_c_hh_TE)
    result_c_lh_TE = array(result_c_lh_TE)
    result_c_lh_TM = 4 * array(result_c_lh_TE)
    result_ex_c_hh_TE = array(result_ex_c_hh_TE)
    result_ex_c_lh_TE = array(result_ex_c_lh_TE)
    result_ex_c_lh_TM = 4 * array(result_ex_c_lh_TE)
    result_TE_raw = result_c_hh_TE + result_c_lh_TE + result_ex_c_hh_TE + result_ex_c_lh_TE
    result_TE = result_TE_raw * (np.cos(theta) ** 2 + 0.5 * np.sin(theta) ** 2)
    result_TM_raw = result_c_lh_TM + result_ex_c_lh_TM
    result_TM = 0.5 * result_TM_raw * np.sin(theta) ** 2
    result_sum = result_TE + result_TM
    return espace, result_sum, result_TE, result_TM, result_TE_raw, result_TM_raw


def NonBlackBodyEmission(E, voltage=0, nr=3.5, T=300):
    """

    :param E:
    :param voltage:
    :param nr:
    :param T:
    :return:
    """
    A = 2 * nr ** 2 / h ** 3 / c ** 2
    B = (E - q * voltage) / kb / T

    bb = A * E ** 2 / (np.exp(B) - 1)

    return bb


def calc_emission(QM_result, well_width, voltage=0, theta=0):
    """

    :param QM_result:
    :param well_width:
    :param voltage:
    :param theta:
    :return:
    """
    alfa = QM_result["alpha"]
    aTE = alfa[4]
    aTM = alfa[5]

    result_TE = []
    result_TM = []
    result = []
    for index, e in enumerate(alfa[0]):
        bb = NonBlackBodyEmission(e)
        TE = bb * (1 - np.exp(-aTE[index] * well_width * (np.cos(theta) ** 2 + 0.5 * np.sin(theta) ** 2)))
        TM = bb * (1 - np.exp(-aTM[index] * well_width * 0.5 * np.sin(theta) ** 2))
        # All = bb*(1-np.exp(-(aTE[index]+aTM[index])*well_width))
        All = TE + TM
        result_TE.append(TE)
        result_TM.append(TM)
        result.append(All)

    result_TE = array(result_TE)
    result_TM = array(result_TM)
    result = array(result)

    return alfa[0], result, result_TE, result_TM



# -*- coding: utf-8 -*-
"""
This is mostly based on Steven Byrnes' tmm package, with modifications to vectorize the code
over wavelengths (by Diego Alonso-Alvarez), and to include depth-dependent absorption calculations
in incoherent layers using the Beer-Lambert law (by Phoebe Pearce).

All credit of the algorithm, testing, etc. goes to Steven Byrnes. For more details, visit:

- https://arxiv.org/abs/1603.02720

- http://sjbyrnes.com

- For information see the docstring of each function, and also see manual.pdf

The most two important functions are:

coh_tmm(...) -- the transfer-matrix-method calculation in the coherent
case (i.e. thin films)

inc_tmm(...) -- the transfer-matrix-method calculation in the incoherent
case (i.e. films tens or hundreds of wavelengths thick, or whose
thickness is not very uniform.

These functions are all imported into the main package (tmm) namespace,
so you can call them with tmm.coh_tmm(...) etc.
"""


import scipy as sp
import numpy as np

import sys

EPSILON = sys.float_info.epsilon  # typical floating-point calculation error


def make_2x2_array(a, b, c, d, dtype=float):
    """
    Makes a 2x2 numpy array of [[a,b],[c,d]]

    Same as "numpy.array([[a,b],[c,d]], dtype=float)", but ten times faster
    """
    my_array = np.empty((len(a), 2, 2), dtype=dtype)
    my_array[:, 0, 0] = a
    my_array[:, 0, 1] = b
    my_array[:, 1, 0] = c
    my_array[:, 1, 1] = d
    return my_array


def snell(n_1, n_2, th_1):
    """
    return angle theta in layer 2 with refractive index n_2, assuming
    it has angle th_1 in layer with refractive index n_1. Use Snell's law. Note
    that "angles" may be complex!!
    """
    # Important that the arcsin here is scipy.arcsin, not numpy.arcsin!! (They
    # give different results e.g. for arcsin(2).)
    # Use real_if_close because e.g. arcsin(2 + 1e-17j) is very different from
    # arcsin(2) due to branch cut
    return sp.arcsin(np.real_if_close(n_1 * np.sin(th_1) / n_2))


def list_snell(n_list, th_0):
    """
    return list of angle theta in each layer based on angle th_0 in layer 0,
    using Snell's law. n_list is index of refraction of each layer. Note that
    "angles" may be complex!!
    """
    # Important that the arcsin here is scipy.arcsin, not numpy.arcsin!! (They
    # give different results e.g. for arcsin(2).)
    # Use real_if_close because e.g. arcsin(2 + 1e-17j) is very different from
    # arcsin(2) due to branch cut
    return sp.arcsin(np.real_if_close(n_list[0] * np.sin(th_0) / n_list))


def interface_r(polarization, n_i, n_f, th_i, th_f):
    """
    reflection amplitude (from Fresnel equations)

    polarization is either "s" or "p" for polarization

    n_i, n_f are (complex) refractive index for incident and final

    th_i, th_f are (complex) propegation angle for incident and final
    (in radians, where 0=normal). "th" stands for "theta".
    """
    if polarization == 's':
        # return 2 * n_i * np.cos(th_i) / (n_i * np.cos(th_i) + n_f * np.cos(th_f))
        return ((n_i * np.cos(th_i) - n_f * np.cos(th_f)) /
                (n_i * np.cos(th_i) + n_f * np.cos(th_f)))
    elif polarization == 'p':
        return ((n_f * np.cos(th_i) - n_i * np.cos(th_f)) /
                (n_f * np.cos(th_i) + n_i * np.cos(th_f)))
    else:
        raise ValueError("Polarization must be 's' or 'p'")


def interface_t(polarization, n_i, n_f, th_i, th_f):
    """
    transmission amplitude (frem Fresnel equations)

    polarization is either "s" or "p" for polarization

    n_i, n_f are (complex) refractive index for incident and final

    th_i, th_f are (complex) propegation angle for incident and final
    (in radians, where 0=normal). "th" stands for "theta".
    """
    if polarization == 's':
        return 2 * n_i * np.cos(th_i) / (n_i * np.cos(th_i) + n_f * np.cos(th_f))
    elif polarization == 'p':
        return 2 * n_i * np.cos(th_i) / (n_f * np.cos(th_i) + n_i * np.cos(th_f))
    else:
        raise ValueError("Polarization must be 's' or 'p'")


def R_from_r(r):
    """
    Calculate reflected power R, starting with reflection amplitude r.
    """
    return abs(r) ** 2


def T_from_t(pol, t, n_i, n_f, th_i, th_f):
    """
    Calculate transmitted power T, starting with transmission amplitude t.

    n_i,n_f are refractive indices of incident and final medium.

    th_i, th_f are (complex) propegation angles through incident & final medium
    (in radians, where 0=normal). "th" stands for "theta".

    In the case that n_i,n_f,th_i,th_f are real, formulas simplify to
    T=|t|^2 * (n_f cos(th_f)) / (n_i cos(th_i)).

    See manual for discussion of formulas
    """
    if (pol == 's'):
        return abs(t ** 2) * (((n_f * np.cos(th_f)).real) / (n_i * np.cos(th_i)).real)
    elif (pol == 'p'):
        return abs(t ** 2) * (((n_f * np.conj(np.cos(th_f))).real) /
                              (n_i * np.conj(np.cos(th_i))).real)
    else:
        raise ValueError("Polarization must be 's' or 'p'")


def power_entering_from_r(pol, r, n_i, th_i):
    """
    Calculate the power entering the first interface of the stack, starting with
    reflection amplitude r. Normally this equals 1-R, but in the unusual case
    that n_i is not real, it can be a bit different than 1-R. See manual.

    n_i is refractive index of incident medium.

    th_i is (complex) propegation angle through incident medium
    (in radians, where 0=normal). "th" stands for "theta".
    """
    if (pol == 's'):
        return ((n_i * np.cos(th_i) * (1 + np.conj(r)) * (1 - r)).real
                / (n_i * np.cos(th_i)).real)
    elif (pol == 'p'):
        return ((n_i * np.conj(np.cos(th_i)) * (1 + r) * (1 - np.conj(r))).real
                / (n_i * np.conj(np.cos(th_i))).real)
    else:
        raise ValueError("Polarization must be 's' or 'p'")


def interface_R(polarization, n_i, n_f, th_i, th_f):
    """
    Fraction of light intensity reflected at an interface.
    """
    r = interface_r(polarization, n_i, n_f, th_i, th_f)
    return R_from_r(r)


def interface_T(polarization, n_i, n_f, th_i, th_f):
    """
    Fraction of light intensity transmitted at an interface.
    """
    t = interface_t(polarization, n_i, n_f, th_i, th_f)
    return T_from_t(polarization, t, n_i, n_f, th_i, th_f)


def coh_tmm(pol, n_list, d_list, th_0, lam_vac):
    """
    This function is vectorized.
    Main "coherent transfer matrix method" calc. Given parameters of a stack,
    calculates everything you could ever want to know about how light
    propagates in it. (If performance is an issue, you can delete some of the
    calculations without affecting the rest.)

    pol is light polarization, "s" or "p".

    n_list is the list of refractive indices, in the order that the light would
    pass through them. The 0'th element of the list should be the semi-infinite
    medium from which the light enters, the last element should be the semi-
    infinite medium to which the light exits (if any exits).

    th_0 is the angle of incidence: 0 for normal, pi/2 for glancing.
    Remember, for a dissipative incoming medium (n_list[0] is not real), th_0
    should be complex so that n0 sin(th0) is real (intensity is constant as
    a function of lateral position).

    d_list is the list of layer thicknesses (front to back). Should correspond
    one-to-one with elements of n_list. First and last elements should be "inf".

    lam_vac is vacuum wavelength of the light.

    Outputs the following as a dictionary (see manual for details)

    * r--reflection amplitude
    * t--transmission amplitude
    * R--reflected wave power (as fraction of incident)
    * T--transmitted wave power (as fraction of incident)
    * power_entering--Power entering the first layer, usually (but not always)
      equal to 1-R (see manual).
    * vw_list-- n'th element is [v_n,w_n], the forward- and backward-traveling
      amplitudes, respectively, in the n'th medium just after interface with
      (n-1)st medium.
    * kz_list--normal component of complex angular wavenumber for
      forward-traveling wave in each layer.
    * th_list--(complex) propagation angle (in radians) in each layer
    * pol, n_list, d_list, th_0, lam_vac--same as input

    """
    # convert lists to numpy arrays if they're not already.
    n_list = np.array(n_list)
    d_list = np.array(d_list, dtype=float)[:, None]

    # input tests
    if hasattr(th_0, 'size') and th_0.size > 1 and th_0.size != lam_vac.size:
        raise ValueError('This function is not vectorized for angles; you need to run one angle calculation at a time.')
    if n_list.shape[0] != d_list.shape[0]:
        raise ValueError("Problem with n_list or d_list!")
    if (d_list[0] != np.inf) or (d_list[-1] != np.inf):
        raise ValueError('d_list must start and end with inf!')
    if any(abs((n_list[0] * np.sin(th_0)).imag) > 100 * EPSILON):
        raise ValueError('Error in n0 or th0!')
    if hasattr(th_0, 'size'):
        th_0 = np.array(th_0)
    num_layers = n_list.shape[0]
    num_wl = n_list.shape[1]

    # th_list is a list with, for each layer, the angle that the light travels
    # through the layer. Computed with Snell's law. Note that the "angles" may be
    # complex!
    th_list = list_snell(n_list, th_0)

    # kz is the z-component of (complex) angular wavevector for forward-moving
    # wave. Positive imaginary part means decaying.
    kz_list = 2 * np.pi * n_list * np.cos(th_list) / lam_vac

    # delta is the total phase accrued by traveling through a given layer.
    # ignore warning about inf multiplication
    olderr = sp.seterr(invalid='ignore')
    delta = kz_list * d_list
    sp.seterr(**olderr)

    # For a very opaque layer, reset delta to avoid divide-by-0 and similar
    # errors. The criterion imag(delta) > 35 corresponds to single-pass
    # transmission < 1e-30 --- small enough that the exact value doesn't
    # matter.
    # It DOES matter (for depth-dependent calculations!)
    delta[1:num_layers - 1, :] = np.where(delta[1:num_layers - 1, :].imag > 100, delta[1:num_layers - 1, :].real + 100j,
                                          delta[1:num_layers - 1, :])

    # t_list[i,j] and r_list[i,j] are transmission and reflection amplitudes,
    # respectively, coming from i, going to j. Only need to calculate this when
    # j=i+1. (2D array is overkill but helps avoid confusion.)
    t_list = np.zeros((num_wl, num_layers, num_layers), dtype=complex)
    r_list = np.zeros((num_wl, num_layers, num_layers), dtype=complex)

    for i in range(num_layers - 1):
        t_list[:, i, i + 1] = interface_t(pol, n_list[i], n_list[i + 1],
                                          th_list[i], th_list[i + 1])
        r_list[:, i, i + 1] = interface_r(pol, n_list[i], n_list[i + 1],
                                          th_list[i], th_list[i + 1])
    # At the interface between the (n-1)st and nth material, let v_n be the
    # amplitude of the wave on the nth side heading forwards (away from the
    # boundary), and let w_n be the amplitude on the nth side heading backwards
    # (towards the boundary). Then (v_n,w_n) = M_n (v_{n+1},w_{n+1}). M_n is
    # M_list[n]. M_0 and M_{num_layers-1} are not defined.
    # My M is a bit different than Sernelius's, but Mtilde is the same.

    M_list = np.zeros((num_layers, num_wl, 2, 2), dtype=complex)
    for i in range(1, num_layers - 1):
        A = make_2x2_array(np.exp(-1j * delta[i]), np.zeros_like(delta[i]), np.zeros_like(delta[i]), np.exp(1j * delta[i]),
                           dtype=complex)
        B = make_2x2_array(np.ones_like(delta[i]), r_list[:, i, i + 1], r_list[:, i, i + 1], np.ones_like(delta[i]),
                           dtype=complex)
        d = (1 / t_list[:, i, i + 1])

        M_list[i] = np.transpose(d * np.transpose(np.matmul(A, B)))  # , (1, 2, 0)), (2, 0, 1))

    Mtilde = make_2x2_array(np.ones_like(delta[i]), np.zeros_like(delta[i]), np.zeros_like(delta[i]),
                            np.ones_like(delta[i]), dtype=complex)
    for i in range(1, num_layers - 1):
        Mtilde = np.matmul(Mtilde, M_list[i])

    A = make_2x2_array(np.ones_like(delta[i]), r_list[:, 0, 1], r_list[:, 0, 1], np.ones_like(delta[i]),
                       dtype=complex)
    d = 1 / t_list[:, 0, 1]
    Mtilde = np.matmul(np.transpose(d * np.transpose(A, (1, 2, 0)), (2, 0, 1)), Mtilde)

    # Net complex transmission and reflection amplitudes
    r = Mtilde[:, 1, 0] / Mtilde[:, 0, 0]
    t = np.ones_like(Mtilde[:, 0, 0]) / Mtilde[:, 0, 0]

    # vw_list[n] = [v_n, w_n]. v_0 and w_0 are undefined because the 0th medium
    # has no left interface.
    vw_list = np.zeros((num_layers, num_wl, 2), dtype=complex)
    vw = np.zeros((num_wl, 2, 2), dtype=complex)
    I = np.identity(2)
    vw[:, 0, 0] = t
    vw[:, 0, 1] = t
    vw_list[-1] = vw[:, 0, :]
    for i in range(num_layers - 2, 0, -1):
        vw = np.matmul(M_list[i], vw)
        vw_list[i, :, :] = vw[:, :, 1]
    vw_list[-1, :, 1] = 0

    # Net transmitted and reflected power, as a proportion of the incoming light
    # power.
    R = R_from_r(r)
    T = T_from_t(pol, t, n_list[0], n_list[-1], th_0, th_list[-1])
    power_entering = power_entering_from_r(
        pol, r, n_list[0], th_0)

    return {'r': r, 't': t, 'R': R, 'T': T, 'power_entering': power_entering,
            'vw_list': vw_list, 'kz_list': kz_list, 'th_list': th_list,
            'pol': pol, 'n_list': n_list, 'd_list': d_list, 'th_0': th_0,
            'lam_vac': lam_vac}

    # return {'r': r, 't': t, 'R': R, 'T': T, 'power_entering': power_entering,
    #         'kz_list': kz_list, 'th_list': th_list,
    #         'pol': pol, 'n_list': n_list, 'd_list': d_list, 'th_0': th_0,
    #         'lam_vac': lam_vac}


def coh_tmm_reverse(pol, n_list, d_list, th_0, lam_vac):
    """
    Reverses the order of the stack then runs coh_tmm.
    """
    th_f = snell(n_list[0], n_list[-1], th_0)
    return coh_tmm(pol, n_list[::-1], d_list[::-1], th_f, lam_vac)


def ellips(n_list, d_list, th_0, lam_vac):
    """
    Calculates ellipsometric parameters, in radians.

    Warning: Conventions differ. You may need to subtract pi/2 or whatever.
    """

    s_data = coh_tmm('s', n_list, d_list, th_0, lam_vac)
    p_data = coh_tmm('p', n_list, d_list, th_0, lam_vac)
    rs = s_data['r']
    rp = p_data['r']
    return {'psi': np.arctan(abs(rp / rs)), 'Delta': np.angle(-rp / rs)}


def unpolarized_RT(n_list, d_list, th_0, lam_vac):
    """
    This function is vectorized.
    Calculates reflected and transmitted power for unpolarized light.
    """

    s_data = coh_tmm('s', n_list, d_list, th_0, lam_vac)
    p_data = coh_tmm('p', n_list, d_list, th_0, lam_vac)
    R = (s_data['R'] + p_data['R']) / 2.
    T = (s_data['T'] + p_data['T']) / 2.
    return {'R': R, 'T': T}


def position_resolved(layer, dist, coh_tmm_data):
    """
    This function is vectorized.
    Starting with output of coh_tmm(), calculate the Poynting vector
    and absorbed energy density a distance "dist" into layer number "layer"
    """

    vw = coh_tmm_data['vw_list'][layer]
    kz = coh_tmm_data['kz_list'][layer]
    th = coh_tmm_data['th_list'][layer]
    n = coh_tmm_data['n_list'][layer]
    n_0 = coh_tmm_data['n_list'][0]
    th_0 = coh_tmm_data['th_0']
    pol = coh_tmm_data['pol']

    # amplitude of forward-moving wave is Ef, backwards is Eb

    Ef = (vw.T[0] * np.exp(1j * kz.T * dist)).T
    Eb = (vw.T[1] * np.exp(-1j * kz.T * dist)).T


    # Poynting vector
    if (pol == 's'):
        poyn = ((n * np.cos(th) * np.conj(Ef + Eb) * (Ef - Eb)).real) / (n_0 * np.cos(th_0)).real
    elif (pol == 'p'):
        poyn = (((n * np.conj(np.cos(th)) * (Ef + Eb) * np.conj(Ef - Eb)).real)
                / (n_0 * np.conj(np.cos(th_0))).real)

    # absorbed energy density
    if (pol == 's'):
        absor = (n * np.cos(th) * kz * abs(Ef + Eb) ** 2).imag / (n_0 * np.cos(th_0)).real
    elif (pol == 'p'):
        absor = (n * np.conj(np.cos(th)) *
                 (kz * abs(Ef - Eb) ** 2 - np.conj(kz) * abs(Ef + Eb) ** 2)
                 ).imag / (n_0 * np.conj(np.cos(th_0))).real


    return ({'poyn': poyn.T, 'absor': absor.T})


def find_in_structure(d_list, dist):
    """
    This function is vectorized.
    d_list is list of thicknesses of layers, all of which are finite.

    dist is the distance from the front of the whole multilayer structure
    (i.e., from the start of layer 0.)

    Function returns [layer,z], where:

    layer is what number layer you're at.
    (For large enough dist, layer = len(d_list), even though d_list[layer]
    doesn't exist in that case.

    z is the distance into that layer.
    """
    if sum(d_list) == np.inf:
        raise ValueError('This function expects finite arguments')
    # layer = 0

    d_array = np.array([0] + d_list)
    cum_sum = d_array.cumsum()
    layer = np.searchsorted(cum_sum, dist, side='right')
    dist = dist - cum_sum[layer - 1]

    return layer, dist


def find_in_structure_with_inf(d_list, dist):
    """
    d_list is list of thicknesses of layers [inf, blah, blah, ..., blah, inf]

    dist is the distance from the front of the whole multilayer structure
    (i.e., frcom the start of layer 1.)

    Function returns [layer,z], where:

    layer is what number layer you're at,

    z is the distance into that layer.
    """
    layer, dist = find_in_structure(d_list[1:-1], dist)
    return layer, dist


def layer_starts(d_list):
    """
    Gives the location of the start of any given layer, relative to the front
    of the whole multilayer structure. (i.e. the start of layer 1)

    d_list is list of thicknesses of layers [inf, blah, blah, ..., blah, inf]

    """
    final_answer = np.zeros(len(d_list))
    final_answer[0] = -np.inf
    final_answer[1] = 0
    for i in range(2, len(d_list)):
        final_answer[i] = final_answer[i - 1] + d_list[i - 1]
    return final_answer


class absorp_analytic_fn:
    """
    This function (specifically, 'run') is vectorized.
    Absorption in a given layer is a pretty simple analytical function:
    The sum of four exponentials.

    a(z) = A1*exp(a1*z) + A2*exp(-a1*z)
           + A3*exp(1j*a3*z) + conj(A3)*exp(-1j*a3*z)

    where a(z) is absorption at depth z, with z=0 being the start of the layer,
    and A1,A2,a1,a3 are real numbers, with a1>0, a3>0, and A3 is complex.
    The class stores these five parameters, as well as d, the layer thickness.

    This gives absorption as a fraction of intensity coming towards the first
    layer of the stack.
    """

    def fill_in(self, coh_tmm_data, layer):
        """
        fill in the absorption analytic function starting from coh_tmm_data
        (the output of coh_tmm), for absorption in the layer with index
        "layer".
        """

        pol = coh_tmm_data['pol']
        v = coh_tmm_data['vw_list'][layer, :, 0]
        w = coh_tmm_data['vw_list'][layer, :, 1]
        kz = coh_tmm_data['kz_list'][layer]
        n = coh_tmm_data['n_list'][layer]
        n_0 = coh_tmm_data['n_list'][0]
        th_0 = coh_tmm_data['th_0']
        th = coh_tmm_data['th_list'][layer]
        self.d = coh_tmm_data['d_list'][layer]

        self.a1 = 2 * kz.imag
        self.a3 = 2 * kz.real

        self.a1[self.a1 < 1e-30] = 0
        self.a3[self.a3 < 1e-30] = 0

        if pol == 's':
            temp = (n * np.cos(th) * kz).imag / (n_0 * np.cos(th_0)).real
            self.A1 = temp * abs(w) ** 2
            self.A2 = temp * abs(v) ** 2
            self.A3 = temp * v * np.conj(w)
        else:  # pol=='p'
            temp = (2 * (kz.imag) * (n * np.cos(np.conj(th))).real /
                    (n_0 * np.conj(np.cos(th_0))).real)
            self.A1 = temp * abs(w) ** 2
            self.A2 = temp * abs(v) ** 2
            self.A3 = v * np.conj(w) * (-2 * (kz.real) * (n * np.cos(np.conj(th))).imag /
                                     (n_0 * np.conj(np.cos(th_0))).real)
        return self

    def copy(self):
        """
        Create copy of an absorp_analytic_fn object
        """
        a = absorp_analytic_fn()
        (a.A1, a.A2, a.A3, a.a1, a.a3, a.d) = (
            self.A1, self.A2, self.A3, self.a1, self.a3, self.d)
        return a

    def run(self, z):
        """
        Calculates absorption at a given depth z, where z=0 is the start of the
        layer.
        """

        if 'ndarray' in str(type(z)) and z.ndim > 0:
            part1 = self.A1[:, None] * np.exp(self.a1[:, None] * z[None, :])
            part2 = self.A2[:, None] * np.exp(-self.a1[:, None] * z[None, :])
            part3 = self.A3[:, None] * np.exp(1j * self.a3[:, None] * z[None, :])
            part4 = np.conj(self.A3[:, None]) * np.exp(-1j * self.a3[:, None] * z[None, :])

            part1[self.A1 < 1e-100, :] = 0
            return (part1 + part2 + part3 + part4)

        else:
            return (self.A1 * np.exp(self.a1 * z) + self.A2 * np.exp(-self.a1 * z)
                    + self.A3 * np.exp(1j * self.a3 * z) + np.conj(self.A3) * np.exp(-1j * self.a3 * z))



    def flip(self):
        """
        Flip the function front-to-back, to describe a(d-z) instead of a(z),
        where d is layer thickness.
        """
        expn = np.exp(self.a1 * self.d)
        #expn[expn > 1e100] = 1e100
        newA1 = self.A2 * np.exp(-self.a1 * self.d)
        newA1[self.A2 == 0] = 0
        newA2 = self.A1 * expn
        newA2[self.A1 == 0] = 0
        self.A1, self.A2 = newA1, newA2
        self.A3 = np.conj(self.A3 * np.exp(1j * self.a3 * self.d))
        return self

    def scale(self, factor):
        """
        multiplies the absorption at each point by "factor".
        """
        self.A1 *= factor
        self.A2 *= factor
        self.A3 *= factor
        self.A1[np.isnan(self.A1)] = 0
        self.A2[np.isnan(self.A2)] = 0
        self.A3[np.isnan(self.A3)] = 0
        return self

    def add(self, b):
        """
        adds another compatible absorption analytical function
        """

        if np.all(b.a1 != self.a1) or np.all(b.a3 != self.a3):
            raise ValueError('Incompatible absorption analytical functions!')

        self.A1 += b.A1
        self.A2 += b.A2
        self.A3 += b.A3
        return self


def absorp_in_each_layer(coh_tmm_data):
    """
    An array listing what proportion of light is absorbed in each layer.

    Assumes the final layer eventually absorbs all transmitted light.

    Assumes the initial layer eventually absorbs all reflected light.

    Entries of array should sum to 1.

    coh_tmm_data is output of coh_tmm()
    """
    num_layers = len(coh_tmm_data['d_list'])
    num_lam_vec = len(coh_tmm_data['lam_vac'])
    power_entering_each_layer = np.zeros((num_layers, num_lam_vec))
    power_entering_each_layer[0] = 1
    power_entering_each_layer[1] = coh_tmm_data['power_entering']
    power_entering_each_layer[-1] = coh_tmm_data['T']
    for i in range(2, num_layers - 1):
        power_entering_each_layer[i] = position_resolved(i, 0, coh_tmm_data)['poyn']
    final_answer = np.zeros((num_layers, num_lam_vec))
    final_answer[0:-1] = -np.diff(power_entering_each_layer, axis=0)
    final_answer[-1] = power_entering_each_layer[-1]


    final_answer[final_answer < 0] = 0

    return final_answer


def inc_group_layers(n_list, d_list, c_list):
    """
    Helper function for inc_tmm. Groups and sorts layer information.

    See coh_tmm for definitions of n_list, d_list.

    c_list is "coherency list". Each entry should be 'i' for incoherent or 'c'
    for 'coherent'.

    A "stack" is a group of one or more consecutive coherent layers. A "stack
    index" labels the stacks 0,1,2,.... The "within-stack index" counts the
    coherent layers within the stack 1,2,3... [index 0 is the incoherent layer
    before the stack starts]

    An "incoherent layer index" labels the incoherent layers 0,1,2,...

    An "alllayer index" labels all layers (all elements of d_list) 0,1,2,...

    Returns info about how the layers relate:

    * stack_d_list[i] = list of thicknesses of each coherent layer in the i'th
      stack, plus starting and ending with "inf"
    * stack_n_list[i] = list of refractive index of each coherent layer in the
      i'th stack, plus the two surrounding incoherent layers
    * all_from_inc[i] = j means that the layer with incoherent index i has
      alllayer index j
    * inc_from_all[i] = j means that the layer with alllayer index i has
      incoherent index j. If j = nan then the layer is coherent.
    * all_from_stack[i1][i2] = j means that the layer with stack index i1 and
      within-stack index i2 has alllayer index j
    * stack_from_all[i] = [j1 j2] means that the layer with alllayer index i is
      part of stack j1 with withinstack-index j2. If stack_from_all[i] = nan
      then the layer is incoherent
    * inc_from_stack[i] = j means that the i'th stack comes after the layer
      with incoherent index j, and before the layer with incoherent index j+1.
    * stack_from_inc[i] = j means that the layer with incoherent index i comes
      immediately after the j'th stack. If j=nan, it is not immediately
      following a stack.
    * num_stacks = number of stacks
    * num_inc_layers = number of incoherent layers
    * num_layers = number of layers total
    """

    if (d_list.ndim != 1):
        raise ValueError("Problem with n_list or d_list!")
    if (d_list[0] != np.inf) or (d_list[-1] != np.inf):
        raise ValueError('d_list must start and end with inf!')
    if (c_list[0] != 'i') or (c_list[-1] != 'i'):
        raise ValueError('c_list should start and end with "i"')
    if not len(n_list) == d_list.size == len(c_list):
        raise ValueError('List sizes do not match!')
    inc_index = 0
    stack_index = 0
    stack_d_list = []
    stack_n_list = []
    all_from_inc = []
    inc_from_all = []
    all_from_stack = []
    stack_from_all = []
    inc_from_stack = []
    stack_from_inc = []
    stack_in_progress = False
    for alllayer_index in range(len(n_list)):
        if c_list[alllayer_index] == 'c':  # coherent layer
            inc_from_all.append(np.nan)
            if not stack_in_progress:  # this layer is starting new stack
                stack_in_progress = True
                ongoing_stack_d_list = [np.inf, d_list[alllayer_index]]
                ongoing_stack_n_list = [n_list[alllayer_index - 1],
                                        n_list[alllayer_index]]
                stack_from_all.append([stack_index, 1])
                all_from_stack.append([alllayer_index - 1, alllayer_index])
                inc_from_stack.append(inc_index - 1)
                within_stack_index = 1
            else:  # another coherent layer in the same stack
                ongoing_stack_d_list.append(d_list[alllayer_index])
                ongoing_stack_n_list.append(n_list[alllayer_index])
                within_stack_index += 1
                stack_from_all.append([stack_index, within_stack_index])
                all_from_stack[-1].append(alllayer_index)
        elif c_list[alllayer_index] == 'i':  # incoherent layer
            stack_from_all.append(np.nan)
            inc_from_all.append(inc_index)
            all_from_inc.append(alllayer_index)
            if not stack_in_progress:  # previous layer was also incoherent
                stack_from_inc.append(np.nan)
            else:  # previous layer was coherent
                stack_in_progress = False
                stack_from_inc.append(stack_index)
                ongoing_stack_d_list.append(np.inf)
                stack_d_list.append(ongoing_stack_d_list)
                ongoing_stack_n_list.append(n_list[alllayer_index])
                stack_n_list.append(ongoing_stack_n_list)
                all_from_stack[-1].append(alllayer_index)
                stack_index += 1
            inc_index += 1
        else:
            raise ValueError("Error: c_list entries must be 'i' or 'c'!")
    return {'stack_d_list': stack_d_list,
            'stack_n_list': stack_n_list,
            'all_from_inc': all_from_inc,
            'inc_from_all': inc_from_all,
            'all_from_stack': all_from_stack,
            'stack_from_all': stack_from_all,
            'inc_from_stack': inc_from_stack,
            'stack_from_inc': stack_from_inc,
            'num_stacks': len(all_from_stack),
            'num_inc_layers': len(all_from_inc),
            'num_layers': len(n_list)}


def inc_tmm(pol, n_list, d_list, c_list, th_0, lam_vac):
    """
    This function is vectorized.
    Incoherent, or partly-incoherent-partly-coherent, transfer matrix method.

    See coh_tmm for definitions of pol, n_list, d_list, th_0, lam_vac.

    c_list is "coherency list". Each entry should be 'i' for incoherent or 'c'
    for 'coherent'.

    If an incoherent layer has real refractive index (no absorption), then its
    thickness doesn't affect the calculation results.

    See https://arxiv.org/abs/1603.02720 for physics background and some
    of the definitions.

    Outputs the following as a dictionary:

    * R--reflected wave power (as fraction of incident)
    * T--transmitted wave power (as fraction of incident)
    * VW_list-- n'th element is [V_n,W_n], the forward- and backward-traveling
      intensities, respectively, at the beginning of the n'th incoherent medium.
    * coh_tmm_data_list--n'th element is coh_tmm_data[n], the output of
      the coh_tmm program for the n'th "stack" (group of one or more
      consecutive coherent layers).
    * coh_tmm_bdata_list--n'th element is coh_tmm_bdata[n], the output of the
      coh_tmm program for the n'th stack, but with the layers of the stack
      in reverse order.
    * stackFB_list--n'th element is [F,B], where F is light traveling forward
      towards the n'th stack and B is light traveling backwards towards the n'th
      stack.
    * num_layers-- total number both coherent and incoherent.
    * power_entering_list--n'th element is the normalized Poynting vector
      crossing the interface into the n'th incoherent layer from the previous
      (coherent or incoherent) layer.
    * Plus, all the outputs of inc_group_layers

    """
    # Convert lists to numpy arrays if they're not already.
    n_list = np.array(n_list)
    d_list = np.array(d_list, dtype=float)

    # Input tests
    # =============================================================================
    #     if hasattr(n_list[0], 'size') and 'complex' not in str(type(n_list[0])):
    #         raise ValueError(
    #             'This function is not vectorized, yet; you need to run one calculation at a time (one wavelength, one angle).')
    # =============================================================================
    if any((np.real_if_close(n_list[0] * np.sin(th_0))).imag != 0):
        raise ValueError('Error in n0 or th0!')

    group_layers_data = inc_group_layers(n_list, d_list, c_list)
    num_inc_layers = group_layers_data['num_inc_layers']
    num_stacks = group_layers_data['num_stacks']
    stack_n_list = group_layers_data['stack_n_list']
    stack_d_list = group_layers_data['stack_d_list']
    all_from_stack = group_layers_data['all_from_stack']
    all_from_inc = group_layers_data['all_from_inc']
    all_from_stack = group_layers_data['all_from_stack']
    stack_from_inc = group_layers_data['stack_from_inc']
    inc_from_stack = group_layers_data['inc_from_stack']

    # th_list is a list with, for each layer, the angle that the light travels
    # through the layer. Computed with Snell's law. Note that the "angles" may be
    # complex!
    th_list = list_snell(n_list, th_0)

    # coh_tmm_data_list[i] is the output of coh_tmm for the i'th stack
    coh_tmm_data_list = []
    # coh_tmm_bdata_list[i] is the same stack as coh_tmm_data_list[i] but
    # with order of layers reversed

    coh_tmm_bdata_list = []
    for i in range(num_stacks):

        coh_tmm_data_list.append(coh_tmm(pol, stack_n_list[i],
                                         stack_d_list[i],
                                         th_list[all_from_stack[i][0]],
                                         lam_vac))
        coh_tmm_bdata_list.append(coh_tmm_reverse(pol, stack_n_list[i],
                                                  stack_d_list[i],
                                                  th_list[all_from_stack[i][0]],
                                                  lam_vac))

    # P_list[i] is fraction not absorbed in a single pass through i'th incoherent
    # layer.

    P_list = np.zeros((num_inc_layers, len(lam_vac)))
    for inc_index in range(1, num_inc_layers - 1):  # skip 0'th and last (infinite)
        i = all_from_inc[inc_index]
        P_list[inc_index] = np.exp(-4 * np.pi * d_list[i]
                                * (n_list[i] * np.cos(th_list[i])).imag / lam_vac)

    # For a very opaque layer, reset P to avoid divide-by-0 and similar
    # errors.
    P_list[P_list < 1e-30] = 1e-30

    # T_list[i,j] and R_list[i,j] are transmission and reflection powers,
    # respectively, coming from the i'th incoherent layer, going to the j'th
    # incoherent layer. Only need to calculate this when j=i+1 or j=i-1.
    # (2D array is overkill but helps avoid confusion.)
    # initialize these arrays
    T_list = np.zeros((num_inc_layers, num_inc_layers, len(lam_vac)))
    R_list = np.zeros((num_inc_layers, num_inc_layers, len(lam_vac)))
    for inc_index in range(num_inc_layers - 1):  # looking at interface i -> i+1
        alllayer_index = all_from_inc[inc_index]
        nextstack_index = stack_from_inc[inc_index + 1]
        if np.isnan(nextstack_index):  # next layer is incoherent
            R_list[inc_index, inc_index + 1] = (
                interface_R(pol, n_list[alllayer_index],
                            n_list[alllayer_index + 1],
                            th_list[alllayer_index],
                            th_list[alllayer_index + 1]))
            T_list[inc_index, inc_index + 1] = (
                interface_T(pol, n_list[alllayer_index],
                            n_list[alllayer_index + 1],
                            th_list[alllayer_index],
                            th_list[alllayer_index + 1]))
            R_list[inc_index + 1, inc_index] = (
                interface_R(pol, n_list[alllayer_index + 1],
                            n_list[alllayer_index],
                            th_list[alllayer_index + 1],
                            th_list[alllayer_index]))
            T_list[inc_index + 1, inc_index] = (
                interface_T(pol, n_list[alllayer_index + 1],
                            n_list[alllayer_index],
                            th_list[alllayer_index + 1],
                            th_list[alllayer_index]))
        else:  # next layer is coherent
            R_list[inc_index, inc_index + 1] = (
                coh_tmm_data_list[nextstack_index]['R'])
            T_list[inc_index, inc_index + 1] = (
                coh_tmm_data_list[nextstack_index]['T'])
            R_list[inc_index + 1, inc_index] = (
                coh_tmm_bdata_list[nextstack_index]['R'])
            T_list[inc_index + 1, inc_index] = (
                coh_tmm_bdata_list[nextstack_index]['T'])

    # L is the transfer matrix from the i'th to (i+1)st incoherent layer, see
    # manual

    # For a very opaque layer, reset T to avoid divide-by-0 and similar
    # errors.
    T_list[T_list < 1e-30] = 1e-30

    L_list = [np.nan]  # L_0 is not defined because 0'th layer has no beginning.
    Ltilde = (np.array([[np.ones_like(lam_vac), -R_list[1, 0]],
                     [R_list[0, 1],
                      T_list[1, 0] * T_list[0, 1] - R_list[1, 0] * R_list[0, 1]]])
              / T_list[0, 1]).transpose(2, 0, 1)

    # Ltilde = Ltilde.transpose(2,0,1)

    for i in range(1, num_inc_layers - 1):
        L = np.matmul(
            np.array([[1 / P_list[i], np.zeros(len(lam_vac))], [np.zeros(len(lam_vac)), P_list[i]]]).transpose(2, 0, 1),
            np.array([[np.ones_like(lam_vac), -R_list[i + 1, i]],
                   [R_list[i, i + 1],
                    T_list[i + 1, i] * T_list[i, i + 1] - R_list[i + 1, i] * R_list[i, i + 1]]]).transpose(2, 0, 1)
        ) / T_list[i, i + 1][:, None, None]
        L_list.append(L)
        Ltilde = np.matmul(Ltilde, L)

    T = 1 / Ltilde[:, 0, 0]
    R = Ltilde[:, 1, 0] / Ltilde[:, 0, 0]

    # VW_list[n] = [V_n, W_n], the forward- and backward-moving intensities
    # at the beginning of the n'th incoherent layer. VW_list[0] is undefined
    # because 0'th layer has no beginning.

    VW_list = np.zeros((num_inc_layers, 2, len(lam_vac)))
    VW_list[0, :, :] = np.nan
    VW = np.array([T, np.zeros(len(lam_vac))])
    VW_list[-1] = VW

    for i in range(num_inc_layers - 2, 0, -1):
        VW = np.matmul(L_list[i], VW_list[i+1].T[:, :, None])
        VW_list[i, :, :] = VW.transpose()


    # stackFB_list[n]=[F,B] means that F is light traveling forward towards n'th
    # stack and B is light traveling backwards towards n'th stack.
    # Reminder: inc_from_stack[i] = j means that the i'th stack comes after the
    # layer with incoherent index j.

    stackFB_list = []
    for stack_index, prev_inc_index in enumerate(inc_from_stack):
        if prev_inc_index == 0:  # stack starts right after semi-infinite layer.
            F = np.ones_like(lam_vac)
        else:
            F = VW_list[prev_inc_index, 0] * P_list[prev_inc_index]
        B = VW_list[prev_inc_index + 1, 1]
        stackFB_list.append([F, B])

    # power_entering_list[i] is the normalized Poynting vector crossing the
    # interface into the i'th incoherent layer from the previous (coherent or
    # incoherent) layer. See manual.
    power_entering_list = [np.ones_like(lam_vac)]  # "1" by convention for infinite 0th layer.
    for i in range(1, num_inc_layers):
        prev_stack_index = stack_from_inc[i]
        if np.isnan(prev_stack_index):
            # case where this layer directly follows another incoherent layer
            if i == 1:  # special case because VW_list[0] & A_list[0] are undefined
                power_entering_list.append(T_list[0, 1]
                                           - VW_list[1, 1] * T_list[1, 0])
            else:
                power_entering_list.append(
                    VW_list[i - 1][0] * P_list[i - 1] * T_list[i - 1, i]
                    - VW_list[i][1] * T_list[i, i - 1])
        else:  # case where this layer follows a coherent stack
            power_entering_list.append(
                stackFB_list[prev_stack_index][0] *
                coh_tmm_data_list[prev_stack_index]['T']
                - stackFB_list[prev_stack_index][1] *
                coh_tmm_bdata_list[prev_stack_index]['power_entering'])

    if len(stackFB_list) > 0:
        stackFB_list_ans = np.stack(stackFB_list).transpose(2, 0, 1)
    else:
        stackFB_list_ans = []
    #('VWlist', VW_list.transpose(2, 0, 1))
    ans = {'T': T, 'R': R, 'VW_list': VW_list.transpose(2, 0, 1),
           'coh_tmm_data_list': coh_tmm_data_list,
           'coh_tmm_bdata_list': coh_tmm_bdata_list,
           'stackFB_list': stackFB_list_ans,
           'power_entering_list': np.stack(power_entering_list).T}
    ans.update(group_layers_data)
    return ans


def inc_absorp_in_each_layer(inc_data):
    """
    A list saying what proportion of light is absorbed in each layer.

    Assumes all reflected light is eventually absorbed in the 0'th medium, and
    all transmitted light is eventually absorbed in the final medium.

    Returns a list [layer0absorp, layer1absorp, ...]. Entries should sum to 1.

    inc_data is output of inc_tmm()
    """
    # Reminder: inc_from_stack[i] = j means that the i'th stack comes after the
    # layer with incoherent index j.
    # Reminder: stack_from_inc[i] = j means that the layer
    # with incoherent index i comes immediately after the j'th stack (or j=nan
    # if it's not immediately following a stack).

    stack_from_inc = inc_data['stack_from_inc']
    power_entering_list = inc_data['power_entering_list'].T
    # stackFB_list[n]=[F,B] means that F is light traveling forward towards n'th
    # stack and B is light traveling backwards towards n'th stack.
    stackFB_list = inc_data['stackFB_list']
    absorp_list = []

    # loop through incoherent layers, excluding the final layer
    for i, power_entering in enumerate(power_entering_list[:-1]):
        if np.isnan(stack_from_inc[i + 1]):
            # case that incoherent layer i is right before another incoherent layer
            absorp_list.append(power_entering_list[i] - power_entering_list[i + 1])
        else:  # incoherent layer i is immediately before a coherent stack
            j = stack_from_inc[i + 1]
            coh_tmm_data = inc_data['coh_tmm_data_list'][j]
            coh_tmm_bdata = inc_data['coh_tmm_bdata_list'][j]
            # First, power in the incoherent layer...
            power_exiting = (
                    stackFB_list[:, j, 0] * coh_tmm_data['power_entering']
                    - stackFB_list[:, j, 1] * coh_tmm_bdata['T'])
            absorp_list.append(power_entering_list[i] - power_exiting)
            # Next, power in the coherent stack...
            stack_absorp = ((stackFB_list[:, j, 0] *
                             absorp_in_each_layer(coh_tmm_data))[1:-1]
                            + (stackFB_list[:, j, 1] *
                               absorp_in_each_layer(coh_tmm_bdata))[-2:0:-1])
            absorp_list.extend(stack_absorp)
    # final semi-infinite layer
    absorp_list.append(inc_data['T'])

    absorp_list = np.array(absorp_list)
    absorp_list[absorp_list < 0] = 0

    return absorp_list


def inc_find_absorp_analytic_fn(layer, inc_data):
    """
    Outputs an absorp_analytic_fn object for a coherent layer within a
    partly-incoherent stack.

    inc_data is output of inc_tmm()
    """
    j = inc_data['stack_from_all'][layer]
    if np.any(np.isnan(j)):
        raise ValueError('layer must be coherent for this function!')
    [stackindex, withinstackindex] = j
    forwardfunc = absorp_analytic_fn()
    forwardfunc.fill_in(inc_data['coh_tmm_data_list'][stackindex],
                        withinstackindex)
    forwardfunc.scale(inc_data['stackFB_list'][:, stackindex, 0])
    backfunc = absorp_analytic_fn()
    backfunc.fill_in(inc_data['coh_tmm_bdata_list'][stackindex],
                     -1 - withinstackindex)

    backfunc.scale(inc_data['stackFB_list'][:, stackindex, 1])

    backfunc.flip()

    return forwardfunc.add(backfunc)


def inc_position_resolved(layer, dist, inc_tmm_data, coherency_list, alphas, zero_threshold=1e-6):
    """
    This function is vectorized. Analogous to position_resolved, but
    for layers (incoherent or coherent) in (partly) incoherent stacks.
    This is a new function, not from Steven Byrnes' tmm package.
    Starting with output of inc_tmm(), calculate the Poynting vector
    and absorbed energy density a distance "dist" into layer number "layer"
    """

    layers = list(set(layer)) # unique layer indices
    A_per_layer = np.array(inc_absorp_in_each_layer(inc_tmm_data))
    fraction_reaching = 1 - np.cumsum(A_per_layer, axis = 0)
    A_local = np.zeros((len(alphas[0]), len(dist)))
    for i, l in enumerate(layers):
        if coherency_list[l] == 'c':

            fn = inc_find_absorp_analytic_fn(l, inc_tmm_data)
            A_layer = fn.run(dist[layer == l])

        else:
            A_layer = beer_lambert(alphas[l] * 1e9, fraction_reaching[i], dist[layer == l] * 1e-9)

        A_layer[fraction_reaching[i] < zero_threshold, :] = 0
        A_local[:, layer == l] = A_layer

    return A_local


def beer_lambert(alphas, fraction, dist):
    """
    Calculates absorption profile according to the Beer-Lambert law given alphas (in m-1)
    and a vector of distance into the layer (in m) and the fraction of incident light
    reaching the front of the layer. This is used to calculate the absorption profile in
    incoherent layers within (partly) incoherent stacks. Vectorized over wavelengths.
    """

    expn = np.exp(- alphas[:, None] * dist[None,:])

    output = fraction[:, None]*alphas[:, None]*expn

    return output/1e9

